"""
Support for GitHub Actions.

GitHub Actions are very flexible, so this code is going to make some
simplifying assumptions:

- you use a matrix strategy
  - on 'python-version' that contains python versions, or
  - on 'config' that contains lists of [python_version, tox_env]
"""

from typing import Optional, Set, Union

import yaml

from .base import Source
from ..parsers.yaml import quote_string, update_yaml_list
from ..sources.tox import toxenv_for_version
from ..utils import FileLines, FileOrFilename, open_file
from ..versions import SortedVersionList, Version


GHA_WORKFLOW_FILE = '.github/workflows/tests.yml'
GHA_WORKFLOW_GLOB = '.github/workflows/*.yml'


def get_gha_python_versions(
    filename: FileOrFilename = GHA_WORKFLOW_FILE,
) -> Optional[SortedVersionList]:
    """Extract supported Python versions from a GitHub workflow."""
    with open_file(filename) as fp:
        conf = yaml.safe_load(fp)

    versions: Set[Version] = set()
    had_matrix = False
    for job_name, job in conf.get('jobs', {}).items():
        matrix = job.get('strategy', {}).get('matrix', {})
        if 'python-version' in matrix:
            had_matrix = True
            versions.update(
                e for e in map(parse_gh_ver, matrix['python-version']) if e)
        if 'config' in matrix:
            had_matrix = True
            versions.update(
                parse_gh_ver(c[0])
                for c in matrix['config']
                if isinstance(c, list)
            )
        if 'include' in matrix:
            for extra in matrix['include']:
                if 'python-version' in extra:
                    had_matrix = True
                    versions.add(parse_gh_ver(extra['python-version']))

    if not had_matrix:
        return None
    return sorted(set(versions))


def parse_gh_ver(v: Union[str, float]) -> Version:
    """Parse Python versions used for actions/setup-python@v2.

    This format is not fully well documented.  There's support for
    specifying things like

    - "3.x" (latest minor in Python 3.x; currently 3.9)
    - "3.7" (latest bugfix in Python 3.7)
    - "3.7.2" (specific version to be downloaded and installed)
    - "pypy2"/"pypy3"
    - "pypy-2.7"/"pypy-3.6"
    - "pypy-3.7-v7.3.3"

    https://github.com/actions/python-versions/blob/main/versions-manifest.json
    contains a list of supported CPython versions that can be downloaded
    and installed; this includes prereleases, but doesn't include PyPy.
    """
    v = str(v)
    if v.startswith(('pypy3', 'pypy-3')):
        return Version.from_string('PyPy3')
    elif v.startswith(('pypy2', 'pypy-2')):
        return Version.from_string('PyPy')
    else:
        return Version.from_string(v)


def update_gha_python_versions(
    filename: FileOrFilename,
    new_versions: SortedVersionList,
) -> FileLines:
    """Update supported Python versions in a GitHub workflow file.

    Does not touch the file but returns a list of lines with new file contents.
    """
    with open_file(filename) as fp:
        orig_lines = fp.readlines()
        fp.seek(0)
        conf = yaml.safe_load(fp)
    new_lines = orig_lines

    def keep_old_version(value: str) -> bool:
        """Determine if a Python version line should be preserved."""
        parsed = yaml.safe_load(value)
        ver = parse_gh_ver(parsed)
        if ver == Version.from_string('PyPy'):
            return any(v.major == 2 for v in new_versions)
        if ver == Version.from_string('PyPy3'):
            return any(v.major == 3 for v in new_versions)
        return False

    def keep_old_config(value: str) -> bool:
        """Determine if a Python version line should be preserved."""
        parsed = yaml.safe_load(value)
        if isinstance(parsed, list) and len(parsed) == 2:
            ver = parse_gh_ver(parsed[0])
            toxenv = str(parsed[1])
        else:
            return True
        if ver == Version.from_string('PyPy'):
            return any(v.major == 2 for v in new_versions)
        if ver == Version.from_string('PyPy3'):
            return any(v.major == 3 for v in new_versions)
        return toxenv != toxenv_for_version(ver)

    for job_name, job in conf.get('jobs', {}).items():
        matrix = job.get('strategy', {}).get('matrix', {})
        if 'python-version' in matrix:
            quote_style = ''
            if all(isinstance(v, str) for v in matrix['python-version']):
                quote_style = '"'
            yaml_new_versions = [
                quote_string(str(v), quote_style)
                for v in new_versions
            ]
            new_lines = update_yaml_list(
                new_lines,
                ('jobs', job_name, 'strategy', 'matrix', 'python-version'),
                yaml_new_versions, filename=fp.name,
                keep=keep_old_version,
            )
        if 'config' in matrix:
            yaml_configs = []
            for v in new_versions:
                quoted_ver = quote_string(str(v), '"')
                toxenv = quote_string(toxenv_for_version(v), '"')
                yaml_configs.append(f"[{quoted_ver + ',':<8} {toxenv}]")
            new_lines = update_yaml_list(
                new_lines,
                ('jobs', job_name, 'strategy', 'matrix', 'config'),
                yaml_configs, filename=fp.name,
                keep=keep_old_config,
            )

    return new_lines


GitHubActions = Source(
    filename=GHA_WORKFLOW_GLOB,
    extract=get_gha_python_versions,
    update=update_gha_python_versions,
    check_pypy_consistency=True,
    has_upper_bound=True,
)



"""
Support for .manylinux-install.sh.  This is a shell script used by multiple
ZopeFoundation packages that builds manylinux wheels inside
quay.io/pypa/manylinux* Docker images.

The script loops over all installed Pythons, checks if each is a supported
version using a series of `if` statements, then builds wheels for each
supported versions.  This looks like ::

    for PYBIN in /opt/python/*/bin; do
        if [[ "${PYBIN}" == *"cp27"* ]] || \
           [[ "${PYBIN}" == *"cp34"* ]] || \
           [[ "${PYBIN}" == *"cp35"* ]] || \
           [[ "${PYBIN}" == *"cp36"* ]] || \
           [[ "${PYBIN}" == *"cp37"* ]]; then
            "${PYBIN}/pip" install -e /io/
            "${PYBIN}/pip" wheel /io/ -w wheelhouse/
            rm -rf /io/build /io/*.egg-info
        fi
    done

"""

import re

from .base import Source
from ..utils import FileLines, FileOrFilename, open_file, warn
from ..versions import SortedVersionList, Version, VersionList


MANYLINUX_INSTALL_SH = '.manylinux-install.sh'


def get_manylinux_python_versions(
    filename: FileOrFilename = MANYLINUX_INSTALL_SH,
) -> SortedVersionList:
    """Extract supported Python versions from .manylinux-install.sh."""
    magic = re.compile(r'.*\[\[ "\$\{PYBIN\}" == \*"cp(\d)(\d+)"\* \]\]')
    versions = []
    with open_file(filename) as fp:
        for line in fp:
            m = magic.match(line)
            if m:
                v = Version.from_string('{}.{}'.format(*m.groups()))
                versions.append(v)
    return sorted(set(versions))


def update_manylinux_python_versions(
    filename: FileOrFilename,
    new_versions: VersionList,
) -> FileLines:
    """Update supported Python versions in .manylinux_install_sh.

    Does not touch the file but returns a list of lines with new file contents.
    """
    magic = re.compile(r'.*\[\[ "\$\{PYBIN\}" == \*"cp(\d)(\d)"\* \]\]')
    with open_file(filename) as f:
        orig_lines = f.readlines()
    lines = iter(enumerate(orig_lines))
    for n, line in lines:
        m = magic.match(line)
        if m:
            start = n
            break
    else:
        warn(f'Failed to understand {f.name}')
        return orig_lines
    for n, line in lines:
        m = magic.match(line)
        if not m:
            end = n
            break
    else:
        warn(f'Failed to understand {f.name}')
        return orig_lines

    indent = ' ' * 4
    conditions = f' || \\\n{indent}   '.join(
        f'[[ "${{PYBIN}}" == *"cp{ver.major}{ver.minor}"* ]]'
        for ver in new_versions
    )
    new_lines = orig_lines[:start] + (
        f'{indent}if {conditions}; then\n'
    ).splitlines(True) + orig_lines[end:]

    return new_lines


Manylinux = Source(
    filename=MANYLINUX_INSTALL_SH,
    extract=get_manylinux_python_versions,
    update=update_manylinux_python_versions,
    check_pypy_consistency=False,
    has_upper_bound=True,
)



"""
Check supported Python versions in a Python package.

Makes sure the set of supported Python versions is consistent between

- setup.py PyPI classifiers
- tox.ini default env list
- .travis-ci.yml
- appveyor.yml
- (optionally) .manylinux-install.sh as used by various ZopeFoundation projects
- .github/workflows/*.yml

"""

__author__ = 'Marius Gedminas <marius@gedmin.as>'
__version__ = '0.19.2'



"""
Command-line user interface.

This is the main module of check-python-versions, responsible for handling
command-line arguments, extracting information about supported Python versions
from various sources, presenting it to the user and possibly making
modifications.
"""

import argparse
import glob
import os
import sys
from io import StringIO
from typing import Callable, Collection, Dict, List, Optional, Tuple

from . import __version__
from .sources.all import ALL_SOURCES
from .sources.base import SourceFile
from .utils import (
    FileLines,
    FileOrFilename,
    confirm_and_update_file,
    show_diff,
)
from .versions import (
    MAX_MINOR_FOR_MAJOR,
    SortedVersionList,
    Version,
    VersionList,
    important,
    pypy_versions,
    update_version_list,
)


def parse_version(v: str) -> Tuple[int, int]:
    """Parse a Python version number.

    Expects 'MAJOR.MINOR', no more, no less.

    Returns a tuple (major, minor).

    This function is used for command-line argument parsing and may raise an
    argparse.ArgumentTypeError.
    """
    try:
        major, minor = map(int, v.split('.', 1))
    except ValueError:
        raise argparse.ArgumentTypeError(f'bad version: {v}')
    return (major, minor)


def parse_version_list(v: str) -> SortedVersionList:
    """Parse a list of Python version ranges.

    Expects something like '2.7,3.6-3.8'.  Allows open ranges.

    Returns an ordered list of strings, each of which represents a single
    version of the form 'MAJOR.MINOR'.

    This function is used for command-line argument parsing and may raise an
    argparse.ArgumentTypeError.
    """
    versions = set()

    for part in v.split(','):
        if '-' in part:
            lo, hi = part.split('-', 1)
        else:
            lo = hi = part

        if lo and hi:
            lo_major, lo_minor = parse_version(lo)
            hi_major, hi_minor = parse_version(hi)
        elif hi and not lo:
            hi_major, hi_minor = parse_version(hi)
            lo_major, lo_minor = hi_major, 0
        elif lo and not hi:
            lo_major, lo_minor = parse_version(lo)
            try:
                hi_major, hi_minor = lo_major, MAX_MINOR_FOR_MAJOR[lo_major]
            except KeyError:
                raise argparse.ArgumentTypeError(
                    f'bad range: {part}')
        else:
            raise argparse.ArgumentTypeError(
                f'bad range: {part}')

        if lo_major != hi_major:
            raise argparse.ArgumentTypeError(
                f'bad range: {part} ({lo_major} != {hi_major})')

        for vmin in range(lo_minor, hi_minor + 1):
            versions.add(Version(major=lo_major, minor=vmin))

    return sorted(versions)


def is_package(where: str = '.') -> bool:
    """Check if there's a Python package in the given directory.

    Currently only traditional packages having a setup.py are supported.

    Does not emit any diagnostics.
    """
    # TODO: support setup.py-less packages that use pyproject.toml instead
    setup_py = os.path.join(where, 'setup.py')
    return os.path.exists(setup_py)


PrintFn = Callable[..., None]


def check_package(where: str = '.', *, print: PrintFn = print) -> bool:
    """Check if there's a Python package in the given directory.

    Currently only traditional packages having a setup.py are supported.

    Emits diagnostics to standard output if ``where`` is not a directory
    or doesn't have a Python package in it.
    """

    if not os.path.isdir(where):
        print("not a directory")
        return False

    if not is_package(where):
        print("no setup.py -- not a Python package?")
        return False

    return True


#
# The way check-python-version does version updates is that it calls
# various update functions and gives them a filename, then gets back
# the updated content as a list of lines.  At the end we can show the diff
# to the user or write them back to the file.
#
# But.  Sometimes we want to call two update functions for the same file
# (setup.py) to update different bits in it (classifiers and python_requires).
# We don't want to write out the result of the first updater to disk before
# we call the second one.  So, here's what we do: we remember the updated
# contents of a file in a "replacement dict", then next time instead of passing
# a filename to an update function we pass it a StringIO() with the intermedate
# results, to get back the final results.
#

ReplacementDict = Dict[str, FileLines]


def filename_or_replacement(
    pathname: str, replacements: Optional[ReplacementDict]
) -> FileOrFilename:
    """Look up a file in the replacement dict.

    This is used to batch multiple updates to a single file.

    Returns the filename if no replacement was found, or a StringIO
    with replacement contents if a replacement was found.
    """
    if replacements and pathname in replacements:
        new_lines = replacements[pathname]
        buf = StringIO("".join(new_lines))
        buf.name = pathname
        return buf
    else:
        return pathname


FilenameSet = Collection[str]


def find_sources(
    where: str = '.',
    *,
    replacements: Optional[ReplacementDict] = None,
    only: Optional[FilenameSet] = None,
    supports_update: bool = False,
) -> List[SourceFile]:
    """Find all sources that exist in a given directory.

    ``replacements`` allows you to check the result of an update (see
    `update_versions`) without actually performing an update.

    ``only`` allows you to check only a subset of the files.

    ``supports_update`` lets you skip sources that don't support updates.
    """
    sources = []
    for source in ALL_SOURCES:
        if supports_update and source.update is None:
            continue  # pragma: nocover
        pathnames = glob.glob(os.path.join(where, source.filename))
        if not pathnames:
            continue
        for pathname in pathnames:
            relpath = os.path.relpath(pathname, where)
            if only and relpath not in only and source.filename not in only:
                continue
            versions = source.extract(
                filename_or_replacement(pathname, replacements))
            if versions is not None:
                sources.append(source.for_file(pathname, versions, relpath))
    return sources


def check_versions(
    where: str = '.',
    *,
    print: PrintFn = print,
    min_width: int = 0,
    expect: Optional[VersionList] = None,
    replacements: Optional[ReplacementDict] = None,
    only: Optional[FilenameSet] = None,
) -> bool:
    """Check Python versions for a single package, located in ``where``.

    ``expect`` allows you to state what versions you expect to be supported.

    ``replacements`` allows you to check the result of an update (see
    `update_versions`) without actually performing an update.

    ``only`` allows you to check only a subset of the files.

    Emits diagnostics to standard output by calling ``print``.
    """

    sources = find_sources(where, replacements=replacements, only=only)

    if not sources:
        print('no file with version information found')
        return False

    width = max(len(source.title) for source in sources) + len(" says:")

    if expect:
        width = max(width, len('expected:'))

    width = max(width, min_width)

    for source in sources:
        print(f"{source.title} says:".ljust(width),
              ", ".join(str(v) for v in source.versions) or "(empty)")

    if expect:
        print("expected:".ljust(width), ', '.join(str(v) for v in expect))

    return supported_versions_match(sources, expect)


def supported_versions_match(
    sources: List[SourceFile],
    expect: Optional[VersionList] = None,
) -> bool:
    version_sets = []
    pypy_version_sets = []

    # This loop covers everything except for setup_requires
    for source in sources:
        if source.has_upper_bound:
            version_sets.append(important(source.versions))
        if source.check_pypy_consistency:
            pypy_version_sets.append(pypy_versions(source.versions))

    # setup_requires usually has no upper bound, which causes trouble when a
    # new Python version gets released.  Let's add an artificial upper bound
    # that matches all the other sources.
    for source in sources:
        if not source.has_upper_bound:
            max_supported_version = max(v for vs in version_sets for v in vs)
            version_sets.append({v for v in important(source.versions)
                                 if v <= max_supported_version})

    if not expect:
        expect = version_sets[0]

    expect = important(expect)
    if not all(expect == v for v in version_sets):
        return False

    if not pypy_version_sets:
        # can't happen: at least one of our sources (setup.py) has pypy info
        return True  # pragma: nocover

    expect_pypy = pypy_version_sets[0]
    return all(expect_pypy == v for v in pypy_version_sets)


def update_versions(
    where: str = '.',
    *,
    add: Optional[VersionList] = None,
    drop: Optional[VersionList] = None,
    update: Optional[VersionList] = None,
    diff: bool = False,
    dry_run: bool = False,
    only: Optional[FilenameSet] = None,
) -> ReplacementDict:
    """Update Python versions for a single package, located in ``where``.

    ``add`` will add to supported versions.
    ``drop`` will remove from supported versions.
    ``update`` will specify supported versions.

    You may combine ``add`` and ``drop``.  It doesn't make sense to combine
    ``update`` with either ``add`` or ``drop``.

    ``only`` allows you to modify only a subset of the files.

    This function performs user interaction: shows a diff, asks for
    confirmation, updates files on disk.

    ``diff``, if true, prints a diff to standard output instead of writing any
    files.

    ``dry_run``, if true, returns a dictionary mapping filenames to new file
    contents instead of asking for confirmation and writing them to disk.
    """

    replacements: ReplacementDict = {}

    sources = find_sources(where, replacements=replacements, only=only,
                           supports_update=True)
    for source in sources:
        # this assert explains supports_update=True to mypy
        assert source.update is not None
        versions = sorted(important(source.versions))
        new_versions = update_version_list(
            versions, add=add, drop=drop, update=update)
        if versions != new_versions:
            fp = filename_or_replacement(source.pathname, replacements)
            new_lines = source.update(fp, new_versions)
            if new_lines is not None:
                # TODO: refactor update_versions() into two functions, one that
                # produces a replacement dict and does no user interaction, and
                # another that does user interaction based on the contents of
                # the replacement dict.  This is because showing a diff for
                # setup.py twice (once to update classifiers and once to update
                # python_requires) is weird?
                if diff:
                    fp = filename_or_replacement(source.pathname, replacements)
                    show_diff(fp, new_lines)
                if dry_run:
                    # XXX: why do this on dry-run only, why not always return a
                    # replacement dict?
                    replacements[source.pathname] = new_lines
                if not diff and not dry_run:
                    confirm_and_update_file(source.pathname, new_lines)

    return replacements


def _main() -> None:
    """The guts of the main() function.

    Parses command-line arguments, does work, reports results, exits with an
    error code if necessary.
    """
    parser = argparse.ArgumentParser(
        description="verify that supported Python versions are the same"
                    " in setup.py, tox.ini, .travis.yml and appveyor.yml")
    parser.add_argument('--version', action='version',
                        version="%(prog)s version " + __version__)
    parser.add_argument('--expect', metavar='VERSIONS',
                        type=parse_version_list,
                        help='expect these versions to be supported, e.g.'
                             ' --expect 2.7,3.5-3.7')
    parser.add_argument('--skip-non-packages', action='store_true',
                        help='skip arguments that are not Python packages'
                             ' without warning about them')
    parser.add_argument('--only', metavar='FILES',
                        help='check only the specified files'
                             ' (comma-separated list, e.g.'
                             ' --only tox.ini,appveyor.yml)')
    parser.add_argument('where', nargs='*',
                        help='directory where a Python package with a setup.py'
                             ' and other files is located')
    group = parser.add_argument_group(
        "updating supported version lists (EXPERIMENTAL)")
    group.add_argument('--add', metavar='VERSIONS', type=parse_version_list,
                       help='add these versions to supported ones, e.g'
                            ' --add 3.8')
    group.add_argument('--drop', metavar='VERSIONS', type=parse_version_list,
                       help='drop these versions from supported ones, e.g'
                            ' --drop 2.6,3.4')
    group.add_argument('--update', metavar='VERSIONS', type=parse_version_list,
                       help='update the set of supported versions, e.g.'
                            ' --update 2.7,3.5-3.7')
    group.add_argument('--diff', action='store_true',
                       help='show a diff of proposed changes')
    group.add_argument('--dry-run', action='store_true',
                       help='verify proposed changes without'
                            ' writing them to disk')
    args = parser.parse_args()

    if args.update and args.add:
        parser.error("argument --add: not allowed with argument --update")
    if args.update and args.drop:
        parser.error("argument --drop: not allowed with argument --update")
    if args.diff and not (args.update or args.add or args.drop):
        parser.error(
            "argument --diff: not allowed without --update/--add/--drop")
    if args.dry_run and not (args.update or args.add or args.drop):
        parser.error(
            "argument --dry-run: not allowed without --update/--add/--drop")
    if args.expect and args.diff and not args.dry_run:
        # XXX: the logic of this escapes me, I think this is because
        # update_versions() doesn't return a replacement dict if you don't use
        # --dry-run?  but why?
        parser.error(
            "argument --expect: not allowed with --diff,"
            " unless you also add --dry-run")

    where = args.where or ['.']
    if args.skip_non_packages:
        where = [path for path in where if is_package(path)]

    only = [a.strip() for a in args.only.split(',')] if args.only else None

    multiple = len(where) > 1

    min_width = 0
    if multiple:
        min_width = max(len(s.title) for s in ALL_SOURCES) + len('says: ')

    mismatches = []
    for n, path in enumerate(where):
        if multiple and (not args.diff or args.dry_run):
            if n:
                print("\n")
            print(f"{path}:\n")
        if not check_package(path):
            mismatches.append(path)
            continue
        replacements = {}
        if args.add or args.drop or args.update:
            replacements = update_versions(
                path, add=args.add, drop=args.drop,
                update=args.update, diff=args.diff,
                dry_run=args.dry_run, only=only)
        if not args.diff or args.dry_run:
            if not check_versions(path, expect=args.expect,
                                  replacements=replacements,
                                  only=only,
                                  min_width=min_width):
                mismatches.append(path)
                continue

    if not args.diff or args.dry_run:
        if mismatches:
            if multiple:
                sys.exit(f"\n\nmismatch in {' '.join(mismatches)}!")
            else:
                sys.exit("\nmismatch!")
        elif multiple:
            print("\n\nall ok!")


def main() -> None:
    """The main function.

    It is here because I detest programs that print tracebacks when they're
    terminated with a Ctrl+C.  I could inline _main() here, but I didn't want
    to indent all of that code.  Maybe I should've added a decorator instead.
    """
    try:
        _main()
    except KeyboardInterrupt:
        sys.exit(2)



"""
Assorted utilities that didn't fit elsewhere.

Yes, this is a sign of bad design.  Maybe someday I'll clean it up.
"""

import difflib
import logging
import os
import stat
import subprocess
import sys
from contextlib import contextmanager
from typing import (
    Any,
    Iterator,
    List,
    Sequence,
    TextIO,
    Tuple,
    TypeVar,
    Union,
    cast,
)


log = logging.getLogger('check-python-versions')


T = TypeVar('T')
OneOrMore = Union[T, Sequence[T]]
OneOrTuple = Union[T, Tuple[T, ...]]


FileObjectWithName = TextIO  # also has a .name attribute
FileOrFilename = Union[str, FileObjectWithName]
FileLines = List[str]


def get_indent(line: str) -> str:
    """Return the indentation part of a line of text."""
    return line[:-len(line.lstrip())]


def warn(msg: str) -> None:
    """Print a warning to standard error."""
    print(msg, file=sys.stderr)


def is_file_object(filename_or_file_object: FileOrFilename) -> bool:
    """Is this a file-like object?"""
    return hasattr(filename_or_file_object, 'read')


@contextmanager
def open_file(filename_or_file_object: FileOrFilename) -> Iterator[TextIO]:
    """Context manager for opening files."""
    if is_file_object(filename_or_file_object):
        yield cast(TextIO, filename_or_file_object)
    else:
        with open(cast(str, filename_or_file_object)) as fp:
            yield fp


def pipe(*cmd: str, **kwargs: Any) -> str:
    """Run a subprocess and return its standard output.

    Keyword arguments are passed directly to `subprocess.Popen`.

    Standard input and standard error are not redirected.
    """
    if 'cwd' in kwargs:
        log.debug('EXEC cd %s && %s', kwargs['cwd'], ' '.join(cmd))
    else:
        log.debug('EXEC %s', ' '.join(cmd))
    p = subprocess.Popen(cmd, stdin=subprocess.DEVNULL, stdout=subprocess.PIPE,
                         **kwargs)
    return cast(bytes, p.communicate()[0]).decode('UTF-8', 'replace')


def confirm_and_update_file(filename: str, new_lines: FileLines) -> None:
    """Update a file with new content, after asking for confirmation."""
    if (show_diff(filename, new_lines)
            and confirm(f"Write changes to {filename}?")):
        mode = stat.S_IMODE(os.stat(filename).st_mode)
        tempfile = filename + '.tmp'
        with open(tempfile, 'w') as f:
            if hasattr(os, 'fchmod'):
                os.fchmod(f.fileno(), mode)
            else:  # pragma: windows
                # Windows, what else?
                os.chmod(tempfile, mode)
            f.writelines(new_lines)
        try:
            os.rename(tempfile, filename)
        except FileExistsError:  # pragma: windows
            # No atomic replace on Windows
            os.unlink(filename)
            os.rename(tempfile, filename)


def show_diff(
    filename_or_file_object: FileOrFilename,
    new_lines: FileLines
) -> bool:
    """Show the difference between two versions of a file."""
    with open_file(filename_or_file_object) as f:
        old_lines = f.readlines()
    print_diff(old_lines, new_lines, f.name)
    return old_lines != new_lines


def print_diff(a: List[str], b: List[str], filename: str) -> None:
    """Show the difference between two versions of a file."""
    print(''.join(difflib.unified_diff(
        a, b,
        filename, filename,
        "(original)", "(updated)",
    )))


def confirm(prompt: str) -> bool:
    """Ask the user to confirm an action."""
    while True:
        try:
            answer = input(f'{prompt} [y/N] ').strip().lower()
        except EOFError:
            answer = ""
        if answer == 'y':
            print()
            return True
        if answer == 'n' or not answer:
            print()
            return False



import pytest

from check_python_versions import versions


@pytest.fixture
def fix_max_python_3_version(monkeypatch):
    def helper(ver):
        monkeypatch.setattr(versions, 'CURRENT_PYTHON_3_VERSION', ver)
        monkeypatch.setitem(versions.MAX_MINOR_FOR_MAJOR, 3, ver)
    return helper



import sys
from io import StringIO

import pytest

from check_python_versions.utils import confirm, get_indent, pipe


@pytest.mark.parametrize(['line', 'expected'], [
    ('no indent\n', ''),
    ('  indent\n', '  '),
    ('  trailing spaces  \n', '  '),
    ('\t  tabs and spaces\n', '\t  '),
    ('\n', ''),
    ('  \n', ''),
    ('', ''),
])
def test_get_indent(line, expected):
    assert get_indent(line) == expected


def test_pipe():
    assert pipe('echo', 'hi') == 'hi\n'


def test_confirm_eof(monkeypatch):
    monkeypatch.setattr(sys, 'stdin', StringIO())
    assert not confirm("Hello how are you?")


def test_confirm_default(monkeypatch):
    monkeypatch.setattr(sys, 'stdin', StringIO("\n"))
    assert not confirm("Hello how are you?")


def test_confirm_no(monkeypatch):
    monkeypatch.setattr(sys, 'stdin', StringIO("n\n"))
    assert not confirm("Hello how are you?")


def test_confirm_yes(monkeypatch):
    monkeypatch.setattr(sys, 'stdin', StringIO("y\n"))
    assert confirm("Hello how are you?")


def test_confirm_neither(monkeypatch):
    monkeypatch.setattr(sys, 'stdin', StringIO("t\ny\n"))
    assert confirm("Hello how are you?")



import pytest

from check_python_versions.versions import (
    Version,
    expand_pypy,
    important,
    pypy_versions,
    update_version_list,
)


def test_version_from_string() -> None:
    assert Version.from_string('3') == Version(major=3)
    assert Version.from_string('3.0') == Version(major=3, minor=0)
    assert Version.from_string('3.6') == Version(major=3, minor=6)
    assert Version.from_string('3.10-dev') == Version('', 3, 10, '-dev')
    assert Version.from_string('PyPy') == Version('PyPy')
    assert Version.from_string('PyPy3') == Version('PyPy', 3)
    assert Version.from_string('PyPy-dev') == Version('PyPy', suffix='-dev')
    assert Version.from_string('PyPy3-dev') == Version('PyPy', 3, -1, '-dev')
    assert Version.from_string('3') != Version.from_string('3.0')


def test_version_repr() -> None:
    assert repr(Version()) == 'Version()'
    assert repr(Version(major=3)) == 'Version(major=3)'
    assert repr(Version(major=3, minor=0)) == 'Version(major=3, minor=0)'
    assert repr(Version(major=3, minor=6)) == 'Version(major=3, minor=6)'
    assert repr(Version(major=3, minor=10, suffix='-dev')) == \
        "Version(major=3, minor=10, suffix='-dev')"
    assert repr(Version(prefix='PyPy')) == "Version(prefix='PyPy')"
    assert repr(Version(prefix='PyPy', major=3, suffix='-dev')) == \
        "Version(prefix='PyPy', major=3, suffix='-dev')"


@pytest.mark.parametrize('v', [
    '3',
    '3.0',
    '3.6',
    '3.10-dev',
    'PyPy',
    'PyPy3',
    'PyPy-dev',
    'PyPy3-dev',
    'PyPy3-dev',
])
def test_version_str_roundtrips(v):
    assert str(Version.from_string(v)) == v


def test_version_sorting():
    assert sorted(Version.from_string(v) for v in [
        '2.7', '3.1', '3.6', '3.10', '3.10-dev',
        'PyPy', 'PyPy3', 'nightly',
    ]) == [Version.from_string(v) for v in [
        '2.7', '3.1', '3.6', '3.10', '3.10-dev',
        'PyPy', 'PyPy3', 'nightly',
    ]]


def test_important(fix_max_python_3_version):
    fix_max_python_3_version(7)
    assert important({
        '2.7', '3.4', '3.7-dev', '3.8', 'nightly', 'PyPy3', 'Jython',
        '3.10.0-beta.3', '3.10.0-alpha.1', '3.10.0-rc.2',
    }) == {'2.7', '3.4'}


def test_pypy_versions():
    assert pypy_versions({
        Version.from_string(v)
        for v in ['2.7', '3.4', '3.7-dev', '3.8', 'nightly', 'PyPy3', 'Jython']
    }) == {Version.from_string(v) for v in ['PyPy3']}


@pytest.mark.parametrize("versions, expected", [
    (['3.6', '2.7'], ['2.7', '3.6']),
    (['2.7', '3.6', 'PyPy'], ['2.7', '3.6', 'PyPy', 'PyPy3']),
    (['2.7', 'PyPy'], ['2.7', 'PyPy']),
    (['3.6', 'PyPy'], ['3.6', 'PyPy3']),
    (['PyPy'], []),  # lol, garbage in, garbage out
])
def test_expand_pypy(versions, expected):
    assert expand_pypy([Version.from_string(v) for v in versions]) == [
        Version.from_string(v) for v in expected
    ]


def test_update_version_list():
    assert update_version_list(['2.7', '3.4']) == ['2.7', '3.4']
    assert update_version_list(['2.7', '3.4'], add=['3.4', '3.5']) == [
        '2.7', '3.4', '3.5',
    ]
    assert update_version_list(['2.7', '3.4'], drop=['3.4', '3.5']) == [
        '2.7',
    ]
    assert update_version_list(['2.7', '3.4'], add=['3.5'], drop=['2.7']) == [
        '3.4', '3.5',
    ]
    assert update_version_list(['2.7', '3.4'], drop=['3.4', '3.5']) == [
        '2.7',
    ]
    assert update_version_list(['2.7', '3.4'], update=['3.4', '3.5']) == [
        '3.4', '3.5',
    ]



"""
Support for setup.py.

There are two ways of declaring Python versions in a setup.py:
classifiers like

    Programming Language :: Python :: 3.8

and python_requires.

check-python-versions supports both.
"""

import ast
import os
import re
import shutil
import sys
from functools import partial
from typing import (
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    TextIO,
    Tuple,
    Union,
    cast,
)

from .base import Source
from ..parsers.python import (
    AstValue,
    eval_ast_node,
    find_call_kwarg_in_ast,
    update_call_arg_in_source,
)
from ..utils import (
    FileLines,
    FileOrFilename,
    is_file_object,
    open_file,
    pipe,
    warn,
)
from ..versions import (
    MAX_MINOR_FOR_MAJOR,
    SortedVersionList,
    Version,
    VersionList,
    expand_pypy,
)


SETUP_PY = 'setup.py'


def get_supported_python_versions(
    filename: FileOrFilename = SETUP_PY
) -> SortedVersionList:
    """Extract supported Python versions from classifiers in setup.py.

    Note: if AST-based parsing fails, this falls back to executing
    ``python setup.py --classifiers``.
    """
    classifiers = get_setup_py_keyword(filename, 'classifiers')
    if classifiers is None and not is_file_object(filename):
        # AST parsing is complicated
        filename = cast(str, filename)
        setup_py = os.path.basename(filename)
        classifiers = pipe(find_python(), setup_py, "-q", "--classifiers",
                           cwd=os.path.dirname(filename)).splitlines()
    if classifiers is None:
        # Note: do not return None because setup.py is not an optional source!
        # We want errors to show up if setup.py fails to declare Python
        # versions in classifiers.
        return []
    if not isinstance(classifiers, (list, tuple)):
        warn('The value passed to setup(classifiers=...) is not a list')
        return []
    return get_versions_from_classifiers(classifiers)


def get_python_requires(
    setup_py: FileOrFilename = SETUP_PY,
) -> Optional[SortedVersionList]:
    """Extract supported Python versions from python_requires in setup.py."""
    python_requires = get_setup_py_keyword(setup_py, 'python_requires')
    if python_requires is None:
        return None
    if not isinstance(python_requires, str):
        warn('The value passed to setup(python_requires=...) is not a string')
        return None
    return parse_python_requires(python_requires)


def is_version_classifier(s: str) -> bool:
    """Is this classifier a Python version classifer?"""
    prefix = 'Programming Language :: Python :: '
    return s.startswith(prefix) and s[len(prefix):len(prefix) + 1].isdigit()


def is_major_version_classifier(s: str) -> bool:
    """Is this classifier a major Python version classifer?

    That is, is this a version classifier that omits the minor version?
    """
    prefix = 'Programming Language :: Python :: '
    return (
        s.startswith(prefix)
        and s[len(prefix):].replace(' :: Only', '').isdigit()
    )


def get_versions_from_classifiers(
    classifiers: Sequence[str],
) -> SortedVersionList:
    """Extract supported Python versions from classifiers."""
    # Based on
    # https://github.com/mgedmin/project-summary/blob/master/summary.py#L221-L234
    prefix = 'Programming Language :: Python :: '
    impl_prefix = 'Programming Language :: Python :: Implementation :: '
    cpython = impl_prefix + 'CPython'
    versions = {
        s[len(prefix):].replace(' :: Only', '').rstrip()
        for s in classifiers
        if is_version_classifier(s)
    } | {
        s[len(impl_prefix):].rstrip()
        for s in classifiers
        if s.startswith(impl_prefix) and s != cpython
    }
    for major in '2', '3':
        if major in versions and any(
                v.startswith(f'{major}.') for v in versions):
            versions.remove(major)
    return expand_pypy(list(map(Version.from_string, versions)))


def update_classifiers(
    classifiers: Sequence[str],
    new_versions: SortedVersionList
) -> List[str]:
    """Update a list of classifiers with new Python versions."""
    prefix = 'Programming Language :: Python :: '

    for pos, s in enumerate(classifiers):
        if is_version_classifier(s):
            break
    else:
        pos = len(classifiers)

    if any(map(is_major_version_classifier, classifiers)):
        new_versions = sorted(
            set(new_versions).union(
                v._replace(prefix='', minor=-1, suffix='')
                for v in new_versions
            )
        )

    classifiers = [
        s for s in classifiers if not is_version_classifier(s)
    ]
    new_classifiers = [
        f'{prefix}{version}'
        for version in new_versions
    ]
    classifiers[pos:pos] = new_classifiers
    return classifiers


def update_supported_python_versions(
    filename: FileOrFilename,
    new_versions: SortedVersionList,
) -> Optional[FileLines]:
    """Update classifiers in a setup.py.

    Does not touch the file but returns a list of lines with new file contents.
    """
    classifiers = get_setup_py_keyword(filename, 'classifiers')
    if classifiers is None:
        return None
    if not isinstance(classifiers, (list, tuple)):
        warn('The value passed to setup(classifiers=...) is not a list')
        return None
    new_classifiers = update_classifiers(classifiers, new_versions)
    return update_setup_py_keyword(filename, 'classifiers', new_classifiers)


def update_python_requires(
    filename: FileOrFilename,
    new_versions: SortedVersionList,
) -> Optional[FileLines]:
    """Update python_requires in a setup.py, if it's defined there.

    Does not touch the file but returns a list of lines with new file contents.
    """
    python_requires = get_setup_py_keyword(filename, 'python_requires')
    if python_requires is None:
        return None
    comma = ', '
    if ',' in python_requires and ', ' not in python_requires:
        comma = ','
    space = ''
    if '> ' in python_requires or '= ' in python_requires:
        space = ' '
    new_python_requires = compute_python_requires(
        new_versions, comma=comma, space=space)
    if is_file_object(filename):
        # Make sure we can read it twice please.
        # XXX: I don't like this.
        cast(TextIO, filename).seek(0)
    return update_setup_py_keyword(filename, 'python_requires',
                                   new_python_requires)


def get_setup_py_keyword(
    setup_py: FileOrFilename,
    keyword: str,
) -> Optional[AstValue]:
    """Extract a value passed to setup() in a setup.py.

    Parses the setup.py into an Abstact Syntax Tree and tries to figure out
    what value was passed to the named keyword argument.

    Returns None if the AST is too complicated to statically evaluate.
    """
    with open_file(setup_py) as f:
        try:
            tree = ast.parse(f.read(), f.name)
        except SyntaxError as error:
            warn(f'Could not parse {f.name}: {error}')
            return None
    node = find_call_kwarg_in_ast(tree, ('setup', 'setuptools.setup'), keyword,
                                  filename=f.name)
    return eval_ast_node(node, keyword) if node is not None else None


def update_setup_py_keyword(
    setup_py: FileOrFilename,
    keyword: str,
    new_value: Union[str, List[str]],
) -> FileLines:
    """Update a value passed to setup() in a setup.py.

    Does not touch the file but returns a list of lines with new file contents.
    """
    with open_file(setup_py) as f:
        lines = f.readlines()
    new_lines = update_call_arg_in_source(lines, ('setup', 'setuptools.setup'),
                                          keyword, new_value)
    return new_lines


def parse_python_requires(s: str) -> Optional[SortedVersionList]:
    """Compute Python versions allowed by a python_requires expression."""

    # https://www.python.org/dev/peps/pep-0440/#version-specifiers
    rx = re.compile(r'^(~=|==|!=|<=|>=|<|>|===)\s*(\d+(?:\.\d+)*(?:\.\*)?)$')

    class BadConstraint(Exception):
        """The version clause is ill-formed according to PEP 440."""

    #
    # This works as follows: we split the specifier on commas into a list
    # of Constraints, each represented as a operator and a tuple of numbers
    # with a possible trailing '*'.  PEP 440 calls them "clauses".
    #

    Constraint = Tuple[Union[str, int], ...]

    #
    # The we look up a handler for each operartor.  This handler takes a
    # constraint and compiles it into a checker.  A checker is a function
    # that takes a Python version number as a 2-tuple and returns True if
    # that version passes its constraint.
    #

    VersionTuple = Tuple[int, int]
    CheckFn = Callable[[VersionTuple], bool]
    HandlerFn = Callable[[Constraint], CheckFn]

    #
    # Here we're defining the handlers for all the operators
    #

    handlers: Dict[str, HandlerFn] = {}
    handler = partial(partial, handlers.__setitem__)

    #
    # We are not doing a strict PEP-440 implementation here because if
    # python_reqiures allows, say, Python 2.7.16, then we want to report that
    # as Python 2.7.  In each handler ``candidate`` is a two-tuple (X, Y)
    # that represents any Python version between X.Y.0 and X.Y.<whatever>.
    #

    @handler('~=')
    def compatible_version(constraint: Constraint) -> CheckFn:
        """~= X.Y more or less means >= X.Y and == X.Y.*"""
        if len(constraint) < 2:
            raise BadConstraint('~= requires a version with at least one dot')
        if constraint[-1] == '*':
            raise BadConstraint('~= does not allow a .*')
        return lambda candidate: candidate == constraint[:2]

    @handler('==')
    def matching_version(constraint: Constraint) -> CheckFn:
        """== X.Y means X.Y, no more, no less; == X[.Y].* is allowed."""
        # we know len(candidate) == 2
        if len(constraint) == 2 and constraint[-1] == '*':
            return lambda candidate: candidate[0] == constraint[0]
        elif len(constraint) == 1:
            # == X should imply Python X.0
            return lambda candidate: candidate == constraint + (0,)
        else:
            # == X.Y.* and == X.Y.Z both imply Python X.Y
            return lambda candidate: candidate == constraint[:2]

    @handler('!=')
    def excluded_version(constraint: Constraint) -> CheckFn:
        """!= X.Y is the opposite of == X.Y."""
        # we know len(candidate) == 2
        if constraint[-1] != '*':
            # != X or != X.Y or != X.Y.Z all are meaningless for us, because
            # there exists some W != Z where we allow X.Y.W and thus allow
            # Python X.Y.
            return lambda candidate: True
        elif len(constraint) == 2:
            # != X.* excludes the entirety of a major version
            return lambda candidate: candidate[0] != constraint[0]
        else:
            # != X.Y.* excludes one particular minor version X.Y,
            # != X.Y.Z.* does not exclude anything, but it's fine,
            # len(candidate) != len(constraint[:-1] so it'll be equivalent to
            # True anyway.
            return lambda candidate: candidate != constraint[:-1]

    @handler('>=')
    def greater_or_equal_version(constraint: Constraint) -> CheckFn:
        """>= X.Y allows X.Y.* or X.(Y+n).*, or (X+n).*."""
        if constraint[-1] == '*':
            raise BadConstraint('>= does not allow a .*')
        # >= X, >= X.Y, >= X.Y.Z all work out nicely because in Python
        # (3, 0) >= (3,)
        return lambda candidate: candidate >= constraint[:2]

    @handler('<=')
    def lesser_or_equal_version(constraint: Constraint) -> CheckFn:
        """<= X.Y is the opposite of > X.Y."""
        if constraint[-1] == '*':
            raise BadConstraint('<= does not allow a .*')
        if len(constraint) == 1:
            # <= X allows up to X.0
            return lambda candidate: candidate <= constraint + (0,)
        else:
            # <= X.Y[.Z] allows up to X.Y
            return lambda candidate: candidate <= constraint

    @handler('>')
    def greater_version(constraint: Constraint) -> CheckFn:
        """> X.Y is equivalent to >= X.Y and != X.Y, I think."""
        if constraint[-1] == '*':
            raise BadConstraint('> does not allow a .*')
        if len(constraint) == 1:
            # > X allows X+1.0 etc
            return lambda candidate: candidate[:1] > constraint
        elif len(constraint) == 2:
            # > X.Y allows X.Y+1 etc
            return lambda candidate: candidate > constraint
        else:
            # > X.Y.Z allows X.Y
            return lambda candidate: candidate >= constraint[:2]

    @handler('<')
    def lesser_version(constraint: Constraint) -> CheckFn:
        """< X.Y is equivalent to <= X.Y and != X.Y, I think."""
        if constraint[-1] == '*':
            raise BadConstraint('< does not allow a .*')
        # < X, < X.Y, < X.Y.Z all work out nicely because in Python
        # (3, 0) > (3,), (3, 0) == (3, 0) and (3, 0) < (3, 0, 1)
        return lambda candidate: candidate < constraint

    @handler('===')
    def arbitrary_version(constraint: Constraint) -> CheckFn:
        """=== X.Y means X.Y, without any zero padding etc."""
        if constraint[-1] == '*':
            raise BadConstraint('=== does not allow a .*')
        # === X does not allow anything
        # === X.Y throws me into confusion; will pip compare Python's X.Y.Z ===
        # X.Y and reject all possible values of Z?
        # === X.Y.Z allows X.Y
        return lambda candidate: candidate == constraint[:2]

    #
    # And now we can do what we planned: split and compile the constraints
    # into checkers (which I also call "constraints", for maximum confusion).
    #

    constraints: List[CheckFn] = []
    for specifier in map(str.strip, s.split(',')):
        m = rx.match(specifier)
        if not m:
            warn(f'Bad python_requires specifier: {specifier}')
            continue
        op, arg = m.groups()
        ver: Constraint = tuple(
            int(segment) if segment != '*' else segment
            for segment in arg.split('.')
        )
        try:
            constraints.append(handlers[op](ver))
        except BadConstraint as error:
            warn(f'Bad python_requires specifier: {specifier} ({error})')

    if not constraints:
        return None

    #
    # And now we can check all the existing Python versions we know about
    # and list those that pass all the requirements.
    #

    versions = []
    for major in sorted(MAX_MINOR_FOR_MAJOR):
        for minor in range(0, MAX_MINOR_FOR_MAJOR[major] + 1):
            if all(constraint((major, minor)) for constraint in constraints):
                versions.append(Version.from_string(f'{major}.{minor}'))
    return versions


def compute_python_requires(
    new_versions: VersionList,
    *,
    comma: str = ', ',
    space: str = '',
) -> str:
    """Compute a value for python_requires that matches a set of versions."""
    new_versions = set(new_versions)
    latest_python = Version(major=3, minor=MAX_MINOR_FOR_MAJOR[3])
    if len(new_versions) == 1 and new_versions != {latest_python}:
        return f'=={space}{new_versions.pop()}.*'
    min_version = min(new_versions)
    specifiers = [f'>={space}{min_version}']
    for major in sorted(MAX_MINOR_FOR_MAJOR):
        for minor in range(0, MAX_MINOR_FOR_MAJOR[major] + 1):
            ver = Version.from_string(f'{major}.{minor}')
            if ver >= min_version and ver not in new_versions:
                specifiers.append(f'!={space}{ver}.*')
    return comma.join(specifiers)


def find_python() -> str:
    """Find a Python interpreter."""
    # The reason I prefer python3 or python from $PATH over sys.executable is
    # this gives the user some control.  E.g. if the setup.py of the project
    # requires some dependencies, the user could install them into a virtualenv
    # and activate it.
    if shutil.which('python3'):
        return 'python3'
    if shutil.which('python'):
        return 'python'
    return sys.executable


SetupClassifiers = Source(
    title=SETUP_PY,
    filename=SETUP_PY,
    extract=get_supported_python_versions,
    update=update_supported_python_versions,
    check_pypy_consistency=True,
    has_upper_bound=True,
)

SetupPythonRequires = Source(
    title='- python_requires',
    filename=SETUP_PY,
    extract=get_python_requires,
    update=update_python_requires,
    check_pypy_consistency=False,
    has_upper_bound=False,  # TBH it might have one!
)



import textwrap
from io import StringIO
from typing import List

from check_python_versions.sources.manylinux import (
    get_manylinux_python_versions,
    update_manylinux_python_versions,
)
from check_python_versions.versions import Version


def v(versions: List[str]) -> List[Version]:
    return [Version.from_string(v) for v in versions]


def test_get_manylinux_python_versions(tmp_path):
    manylinux_install_sh = tmp_path / ".manylinux-install.sh"
    manylinux_install_sh.write_text(textwrap.dedent(r"""
        #!/usr/bin/env bash

        set -e -x

        # Compile wheels
        for PYBIN in /opt/python/*/bin; do
            if [[ "${PYBIN}" == *"cp27"* ]] || \
               [[ "${PYBIN}" == *"cp34"* ]] || \
               [[ "${PYBIN}" == *"cp35"* ]] || \
               [[ "${PYBIN}" == *"cp36"* ]] || \
               [[ "${PYBIN}" == *"cp310"* ]] || \
               [[ "${PYBIN}" == *"cp37"* ]]; then
                "${PYBIN}/pip" install -e /io/
                "${PYBIN}/pip" wheel /io/ -w wheelhouse/
                   rm -rf /io/build /io/*.egg-info
            fi
        done

        # Bundle external shared libraries into the wheels
        for whl in wheelhouse/zope.interface*.whl; do
            auditwheel repair "$whl" -w /io/wheelhouse/
        done
    """.lstrip('\n')))
    assert get_manylinux_python_versions(manylinux_install_sh) == v([
        '2.7', '3.4', '3.5', '3.6', '3.7', '3.10',
    ])


def test_update_manylinux_python_versions():
    manylinux_install_sh = StringIO(textwrap.dedent(r"""
        #!/usr/bin/env bash

        set -e -x

        # Compile wheels
        for PYBIN in /opt/python/*/bin; do
            if [[ "${PYBIN}" == *"cp27"* ]] || \
               [[ "${PYBIN}" == *"cp34"* ]] || \
               [[ "${PYBIN}" == *"cp35"* ]] || \
               [[ "${PYBIN}" == *"cp36"* ]] || \
               [[ "${PYBIN}" == *"cp37"* ]]; then
                "${PYBIN}/pip" install -e /io/
                "${PYBIN}/pip" wheel /io/ -w wheelhouse/
                   rm -rf /io/build /io/*.egg-info
            fi
        done

        # Bundle external shared libraries into the wheels
        for whl in wheelhouse/zope.interface*.whl; do
            auditwheel repair "$whl" -w /io/wheelhouse/
        done
    """).lstrip('\n'))
    result = update_manylinux_python_versions(
        manylinux_install_sh,
        v(['3.6', '3.7', '3.8', '3.10']),
    )
    # NB: when Python 3.10 arrives, any packages still declaring support for
    # Python 3.1 will be problematic, because "cp310" matches *"cp31"*.
    # Luckily no packages using manyinux-install.sh and supporting Python 3.1
    # exist any more.
    assert "".join(result) == textwrap.dedent(r"""
        #!/usr/bin/env bash

        set -e -x

        # Compile wheels
        for PYBIN in /opt/python/*/bin; do
            if [[ "${PYBIN}" == *"cp36"* ]] || \
               [[ "${PYBIN}" == *"cp37"* ]] || \
               [[ "${PYBIN}" == *"cp38"* ]] || \
               [[ "${PYBIN}" == *"cp310"* ]]; then
                "${PYBIN}/pip" install -e /io/
                "${PYBIN}/pip" wheel /io/ -w wheelhouse/
                   rm -rf /io/build /io/*.egg-info
            fi
        done

        # Bundle external shared libraries into the wheels
        for whl in wheelhouse/zope.interface*.whl; do
            auditwheel repair "$whl" -w /io/wheelhouse/
        done
    """).lstrip('\n')


def test_update_manylinux_python_versions_failure(capsys):
    manylinux_install_sh = StringIO(textwrap.dedent(r"""
        #!/usr/bin/env bash

        # TBD
    """).lstrip('\n'))
    manylinux_install_sh.name = '.manylinux-install.sh'
    result = update_manylinux_python_versions(
        manylinux_install_sh,
        v(['3.6', '3.7', '3.8']),
    )
    assert "".join(result) == textwrap.dedent(r"""
        #!/usr/bin/env bash

        # TBD
    """).lstrip('\n')
    assert (
        "Failed to understand .manylinux-install.sh"
        in capsys.readouterr().err
    )


def test_update_manylinux_python_versions_truncated(capsys):
    manylinux_install_sh = StringIO(textwrap.dedent(r"""
        #!/usr/bin/env bash

        set -e -x

        # Compile wheels
        for PYBIN in /opt/python/*/bin; do
            if [[ "${PYBIN}" == *"cp27"* ]] || \
               [[ "${PYBIN}" == *"cp34"* ]] || \
               [[ "${PYBIN}" == *"cp35"* ]] || \
    """).lstrip('\n'))
    manylinux_install_sh.name = '.manylinux-install.sh'
    result = update_manylinux_python_versions(
        manylinux_install_sh,
        v(['3.6', '3.7', '3.8']),
    )
    assert "".join(result) == textwrap.dedent(r"""
        #!/usr/bin/env bash

        set -e -x

        # Compile wheels
        for PYBIN in /opt/python/*/bin; do
            if [[ "${PYBIN}" == *"cp27"* ]] || \
               [[ "${PYBIN}" == *"cp34"* ]] || \
               [[ "${PYBIN}" == *"cp35"* ]] || \
    """).lstrip('\n')
    assert (
        "Failed to understand .manylinux-install.sh"
        in capsys.readouterr().err
    )



import textwrap
from io import StringIO
from typing import List

import pytest

from check_python_versions.sources.travis import (
    XENIAL_SUPPORTED_PYPY_VERSIONS,
    get_travis_yml_python_versions,
    travis_normalize_py_version,
    update_travis_yml_python_versions,
)
from check_python_versions.versions import Version


def v(versions: List[str]) -> List[Version]:
    return [Version.from_string(v) for v in versions]


def test_get_travis_yml_python_versions(tmp_path):
    travis_yml = tmp_path / ".travis.yml"
    travis_yml.write_text(textwrap.dedent("""\
        python:
          - 2.7
          - 3.6
          - 3.10-dev
        matrix:
          include:
            - python: 3.7
            - name: something unrelated
        jobs:
          include:
            - python: 3.4
            - name: something unrelated
        env:
          - TOXENV=py35-docs
          - UNRELATED=variable
    """))
    assert get_travis_yml_python_versions(travis_yml) == v([
        '2.7', '3.4', '3.5', '3.6', '3.7', '3.10-dev',
    ])


def test_get_travis_yml_python_versions_no_list(tmp_path):
    travis_yml = StringIO(textwrap.dedent("""\
        python: 3.7
    """))
    travis_yml.name = '.travis.yml'
    assert get_travis_yml_python_versions(travis_yml) == v([
        '3.7',
    ])


def test_get_travis_yml_python_versions_no_python_only_matrix(tmp_path):
    travis_yml = tmp_path / ".travis.yml"
    travis_yml.write_text(textwrap.dedent("""\
        matrix:
          include:
            - python: 3.7
    """))
    assert get_travis_yml_python_versions(travis_yml) == v([
        '3.7',
    ])


@pytest.mark.parametrize('s, expected', [
    (3.6, '3.6'),
    ('3.7', '3.7'),
    ('pypy', 'PyPy'),
    ('pypy2', 'PyPy'),
    ('pypy2.7', 'PyPy'),
    ('pypy2.7-5.10.0', 'PyPy'),
    ('pypy3', 'PyPy3'),
    ('pypy3.5', 'PyPy3'),
    ('pypy3.5-5.10.1', 'PyPy3'),
    ('3.7-dev', '3.7-dev'),
    ('nightly', 'nightly'),
])
def test_travis_normalize_py_version(s, expected):
    assert travis_normalize_py_version(s) == Version.from_string(expected)


def test_update_travis_yml_python_versions():
    travis_yml = StringIO(textwrap.dedent("""\
        language: python
        python:
          - 2.7
          - pypy
        install: pip install -e .
        script: pytest tests
    """))
    travis_yml.name = '.travis.yml'
    result = update_travis_yml_python_versions(travis_yml, v(["2.7", "3.4"]))
    assert "".join(result) == textwrap.dedent("""\
        language: python
        python:
          - 2.7
          - 3.4
          - pypy
        install: pip install -e .
        script: pytest tests
    """)


def test_update_travis_yml_python_versions_3_10():
    travis_yml = StringIO(textwrap.dedent("""\
        language: python
        python:
          - 3.4
          - pypy3
        install: pip install -e .
        script: pytest tests
    """))
    travis_yml.name = '.travis.yml'
    result = update_travis_yml_python_versions(travis_yml, v(["3.9", "3.10"]))
    # Note: we cannot say '- 3.10', that's a float and evaluates to 3.1!
    assert "".join(result) == textwrap.dedent("""\
        language: python
        python:
          - 3.9
          - "3.10"
          - pypy3
        install: pip install -e .
        script: pytest tests
    """)


def test_update_travis_yml_python_versions_as_strings():
    travis_yml = StringIO(textwrap.dedent("""\
        language: python
        python:
          - "3.4"
          - pypy3
        install: pip install -e .
        script: pytest tests
    """))
    travis_yml.name = '.travis.yml'
    result = update_travis_yml_python_versions(travis_yml, v(["3.9"]))
    assert "".join(result) == textwrap.dedent("""\
        language: python
        python:
          - "3.9"
          - pypy3
        install: pip install -e .
        script: pytest tests
    """)


def test_update_travis_yml_some_python_versions_as_strings():
    travis_yml = StringIO(textwrap.dedent("""\
        language: python
        python:
          - 3.9
          - "3.10"
          - pypy3
        install: pip install -e .
        script: pytest tests
    """))
    travis_yml.name = '.travis.yml'
    result = update_travis_yml_python_versions(travis_yml, v(["3.9", "3.10"]))
    assert "".join(result) == textwrap.dedent("""\
        language: python
        python:
          - 3.9
          - "3.10"
          - pypy3
        install: pip install -e .
        script: pytest tests
    """)


def test_update_travis_yml_python_versions_drops_pypy():
    travis_yml = StringIO(textwrap.dedent("""\
        language: python
        python:
          - 2.7
          - 3.4
          - pypy
          - pypy3
        install: pip install -e .
        script: pytest tests
    """))
    travis_yml.name = '.travis.yml'
    result = update_travis_yml_python_versions(travis_yml, v(["3.8"]))
    assert "".join(result) == textwrap.dedent("""\
        language: python
        python:
          - 3.8
          - pypy3
        install: pip install -e .
        script: pytest tests
    """)


def test_update_travis_yml_python_versions_drops_pypy3():
    # yes this test case is massively unrealistic
    travis_yml = StringIO(textwrap.dedent("""\
        language: python
        python:
          - 2.7
          - 3.4
          - pypy
          - pypy3
        install: pip install -e .
        script: pytest tests
    """))
    travis_yml.name = '.travis.yml'
    result = update_travis_yml_python_versions(travis_yml, v(["2.7"]))
    assert "".join(result) == textwrap.dedent("""\
        language: python
        python:
          - 2.7
          - pypy
        install: pip install -e .
        script: pytest tests
     """)


def test_update_travis_yml_python_versions_keeps_dev():
    travis_yml = StringIO(textwrap.dedent("""\
        language: python
        python:
          - 3.7
          - 3.8
          - 3.9-dev
        install: pip install -e .
        script: pytest tests
    """))
    travis_yml.name = '.travis.yml'
    result = update_travis_yml_python_versions(travis_yml, v(["3.8"]))
    assert "".join(result) == textwrap.dedent("""\
        language: python
        python:
          - 3.8
          - 3.9-dev
        install: pip install -e .
        script: pytest tests
    """)


def test_update_travis_yml_python_versions_drops_dist_trusty(monkeypatch):
    monkeypatch.setitem(
        XENIAL_SUPPORTED_PYPY_VERSIONS, 'pypy', 'pypy2.7-6.0.0')
    travis_yml = StringIO(textwrap.dedent("""\
        language: python
        dist: trusty
        python:
          - 2.7
          - pypy
        install: pip install -e .
        script: pytest tests
    """))
    travis_yml.name = '.travis.yml'
    result = update_travis_yml_python_versions(travis_yml, v(["2.7", "3.7"]))
    assert "".join(result) == textwrap.dedent("""\
        language: python
        python:
          - 2.7
          - 3.7
          - pypy2.7-6.0.0
        install: pip install -e .
        script: pytest tests
    """)


def test_update_travis_yml_python_versions_drops_sudo():
    travis_yml = StringIO(textwrap.dedent("""\
        language: python
        sudo: false
        dist: xenial
        python:
          - 2.7
        install: pip install -e .
        script: pytest tests
    """))
    travis_yml.name = '.travis.yml'
    result = update_travis_yml_python_versions(travis_yml, v(["2.7", "3.7"]))
    assert "".join(result) == textwrap.dedent("""\
        language: python
        dist: xenial
        python:
          - 2.7
          - 3.7
        install: pip install -e .
        script: pytest tests
    """)


def test_update_travis_yml_python_versions_drops_matrix():
    travis_yml = StringIO(textwrap.dedent("""\
        language: python
        python:
          - 2.6
          - 2.7
        matrix:
          include:
            - python: 3.7
              sudo: required
              dist: xenial
        install: pip install -e .
        script: pytest tests
    """))
    travis_yml.name = '.travis.yml'
    result = update_travis_yml_python_versions(travis_yml, v(["2.7", "3.7"]))
    assert "".join(result) == textwrap.dedent("""\
        language: python
        python:
          - 2.7
          - 3.7
        install: pip install -e .
        script: pytest tests
    """)


def test_update_travis_yml_python_versions_keeps_matrix():
    travis_yml = StringIO(textwrap.dedent("""\
        language: python
        python:
          - 2.7
        matrix:
          include:
            - python: 2.7
              env: MINIMAL=1
        install: pip install -e .
        script: pytest tests
    """))
    travis_yml.name = '.travis.yml'
    result = update_travis_yml_python_versions(travis_yml, v(["2.7", "3.7"]))
    assert "".join(result) == textwrap.dedent("""\
        language: python
        python:
          - 2.7
          - 3.7
        matrix:
          include:
            - python: 2.7
              env: MINIMAL=1
        install: pip install -e .
        script: pytest tests
    """)


def test_update_travis_yml_python_versions_one_to_many():
    travis_yml = StringIO(textwrap.dedent("""\
        language: python
        python: 2.7
        install: pip install -e .
        script: pytest tests
    """))
    travis_yml.name = '.travis.yml'
    result = update_travis_yml_python_versions(travis_yml, v(["2.7", "3.4"]))
    assert "".join(result) == textwrap.dedent("""\
        language: python
        python:
          - 2.7
          - 3.4
        install: pip install -e .
        script: pytest tests
    """)


def test_update_travis_yml_python_versions_matrix():
    travis_yml = StringIO(textwrap.dedent("""\
        language: python
        matrix:
          exclude:
            - python: 2.6
          # this is where the fun begins!
          include:
            - python: 2.7
            - python: 3.3
            - python: pypy
            - name: docs
              python: 2.7
              install: pip install sphinx
              script: sphinx-build .
            - name: flake8
              python: 2.7
              install: pip install flake8
              script: flake8 .
        install: pip install -e .
        script: pytest tests
    """))
    travis_yml.name = '.travis.yml'
    result = update_travis_yml_python_versions(travis_yml, v(["2.7", "3.4"]))
    assert "".join(result) == textwrap.dedent("""\
        language: python
        matrix:
          exclude:
            - python: 2.6
          # this is where the fun begins!
          include:
            - python: 2.7
            - python: 3.4
            - python: pypy
            - name: docs
              python: 2.7
              install: pip install sphinx
              script: sphinx-build .
            - name: flake8
              python: 2.7
              install: pip install flake8
              script: flake8 .
        install: pip install -e .
        script: pytest tests
    """)


def test_update_travis_yml_python_versions_matrix_xenial(monkeypatch):
    monkeypatch.setitem(
        XENIAL_SUPPORTED_PYPY_VERSIONS, 'pypy', 'pypy2.7-6.0.0')
    travis_yml = StringIO(textwrap.dedent("""\
        language: python
        matrix:
          exclude:
            - python: 2.6
          # this is where the fun begins!
          include:
            - python: 2.7
            - python: 3.3
            - python: pypy
            - name: docs
              python: 2.7
              install: pip install sphinx
              script: sphinx-build .
            - name: flake8
              python: 2.7
              install: pip install flake8
              script: flake8 .
        install: pip install -e .
        script: pytest tests
    """))
    travis_yml.name = '.travis.yml'
    result = update_travis_yml_python_versions(travis_yml, v(["2.7", "3.7"]))
    assert "".join(result) == textwrap.dedent("""\
        language: python
        matrix:
          exclude:
            - python: 2.6
          # this is where the fun begins!
          include:
            - python: 2.7
            - python: 3.7
            - python: pypy2.7-6.0.0
            - name: docs
              python: 2.7
              install: pip install sphinx
              script: sphinx-build .
            - name: flake8
              python: 2.7
              install: pip install flake8
              script: flake8 .
        install: pip install -e .
        script: pytest tests
    """)


def test_update_travis_yml_python_versions_matrix_preserve_quotes():
    travis_yml = StringIO(textwrap.dedent("""\
        language: python
        matrix:
          include:
            - python: "2.7"
            - python: "3.3"
            - language: c
        install: pip install tox
        script: tox -e py
    """))
    travis_yml.name = '.travis.yml'
    result = update_travis_yml_python_versions(travis_yml, v(["3.8", "3.9"]))
    assert "".join(result) == textwrap.dedent("""\
        language: python
        matrix:
          include:
            - python: "3.8"
            - python: "3.9"
            - language: c
        install: pip install tox
        script: tox -e py
    """)



import textwrap

from check_python_versions.parsers.ini import update_ini_setting


def test_update_ini_setting():
    source_lines = textwrap.dedent("""\
        [tox]
        envlist = py26,py27
        usedevelop = true
    """).splitlines(True)
    result = update_ini_setting(source_lines, 'tox', 'envlist', 'py36,py37',
                                filename='tox.ini')
    assert "".join(result) == textwrap.dedent("""\
        [tox]
        envlist = py36,py37
        usedevelop = true
    """)


def test_update_ini_setting_nospaces():
    source_lines = textwrap.dedent("""\
        [tox]
        envlist=py26,py27
        usedevelop=true
    """).splitlines(True)
    result = update_ini_setting(source_lines, 'tox', 'envlist', 'py36,py37',
                                filename='tox.ini')
    assert "".join(result) == textwrap.dedent("""\
        [tox]
        envlist=py36,py37
        usedevelop=true
    """)


def test_update_ini_setting_from_empty():
    source_lines = textwrap.dedent("""\
        [tox]
        envlist =
        usedevelop = true
    """).splitlines(True)
    result = update_ini_setting(source_lines, 'tox', 'envlist', 'py36,py37',
                                filename='tox.ini')
    assert "".join(result) == textwrap.dedent("""\
        [tox]
        envlist = py36,py37
        usedevelop = true
    """)


def test_update_ini_setting_multiline():
    source_lines = textwrap.dedent("""\
        [tox]
        envlist =
            py26,
            py27
        usedevelop = true
    """).splitlines(True)
    result = update_ini_setting(source_lines, 'tox', 'envlist', 'py36,\npy37',
                                filename='tox.ini')
    assert "".join(result) == textwrap.dedent("""\
        [tox]
        envlist =
            py36,
            py37
        usedevelop = true
    """)


def test_update_ini_setting_multiline_first_on_same_line():
    source_lines = textwrap.dedent("""\
        [tox]
        envlist = py26,
                  py27
        usedevelop = true
    """).splitlines(True)
    result = update_ini_setting(source_lines, 'tox', 'envlist', 'py36,\npy37',
                                filename='tox.ini')
    assert "".join(result) == textwrap.dedent("""\
        [tox]
        envlist = py36,
                  py37
        usedevelop = true
    """)


def test_update_ini_setting_multiline_with_comments():
    source_lines = textwrap.dedent("""\
        [tox]
        envlist =
        # blah blah
        #   py26,py27,pypy
            py26,py27
        # etc.
        usedevelop = true
    """).splitlines(True)
    result = update_ini_setting(source_lines, 'tox', 'envlist', 'py36,py37',
                                filename='tox.ini')
    assert "".join(result) == textwrap.dedent("""\
        [tox]
        envlist =
        # blah blah
        #   py26,py27,pypy
            py36,py37
        # etc.
        usedevelop = true
    """)


def test_update_ini_setting_no_section(capsys):
    source_lines = textwrap.dedent("""\
        [toxx]
    """).splitlines(True)
    result = update_ini_setting(source_lines, 'tox', 'envlist', 'py36,py37',
                                filename='tox.ini')
    assert "".join(result) == textwrap.dedent("""\
        [toxx]
    """)
    assert (
        "Did not find [tox] section in tox.ini"
        in capsys.readouterr().err
    )


def test_update_ini_setting_no_key(capsys):
    source_lines = textwrap.dedent("""\
        [tox]
        usedevelop = true
    """).splitlines(True)
    result = update_ini_setting(source_lines, 'tox', 'envlist', 'py36,py37',
                                filename='tox.ini')
    assert "".join(result) == textwrap.dedent("""\
        [tox]
        usedevelop = true
    """)
    assert (
        "Did not find envlist= in [tox] in tox.ini"
        in capsys.readouterr().err
    )



"""Python version business logic."""

import re
from typing import Collection, List, NamedTuple, Optional, Set, Union


#
# Information about Python releases that needs to be constantly updated as
# Python makes new releases.
#

MAX_PYTHON_1_VERSION = 6  # i.e. 1.6
MAX_PYTHON_2_VERSION = 7  # i.e. 2.7
CURRENT_PYTHON_3_VERSION = 10  # i.e. 3.10

MAX_MINOR_FOR_MAJOR = {
    1: MAX_PYTHON_1_VERSION,
    2: MAX_PYTHON_2_VERSION,
    3: CURRENT_PYTHON_3_VERSION,
}


VERSION_RX = re.compile('^([^-0-9]*)([0-9]*)([.][0-9]+)?(.*)$')


class Version(NamedTuple):
    """A simplified Python version number.

    Primarily needed so we can sort lists of version numbers correctly, i.e.
    2.7, 3.0, 3.1, 3.2, ..., 3.9, 3.10, ...

    Can have an optional prefix, e.g. PyPy3.6 is Version(prefix='PyPy',
    major=3, minor=6).

    Can have an optional suffix, e.g. 3.10-dev is Version(major=3, minor=10,
    suffix='-dev').

    Any string can be round-tripped to a Version and back via
    Version.from_string() and Version.__str__.
    """

    prefix: str = ''
    major: int = -1  # I'd've preferred to use None, but it complicates sorting
    minor: int = -1
    suffix: str = ''

    @classmethod
    def from_string(cls, v: str) -> 'Version':
        m = VERSION_RX.match(v)
        assert m is not None
        prefix, major, minor, suffix = m.groups()
        return cls(
            prefix,
            int(major) if major else -1,
            int(minor[1:]) if minor else -1,
            suffix,
        )

    def __repr__(self) -> str:
        return 'Version({})'.format(', '.join(part for part in [
            f'prefix={self.prefix!r}' if self.prefix else '',
            f'major={self.major!r}' if self.major != -1 else '',
            f'minor={self.minor!r}' if self.minor != -1 else '',
            f'suffix={self.suffix!r}' if self.suffix else '',
        ] if part))

    def __str__(self) -> str:
        major = '' if self.major == -1 else f'{self.major}'
        minor = '' if self.minor == -1 else f'.{self.minor}'
        return f'{self.prefix}{major}{minor}{self.suffix}'


VersionSet = Set[Version]
VersionList = Collection[Version]
SortedVersionList = List[Version]


def is_important(v: Union[Version, str]) -> bool:
    """Is the version important for matching purposes?

    Different sources can express support for different versions, e.g.
    classifiers can express support for "PyPy" but python_requires can't.
    Also some CI systems allow testing on unreleased Python versions that
    cannot be listed in classifiers, so their presence should not cause
    mismatch errors.
    """
    if not isinstance(v, Version):
        v = Version.from_string(v)
    upcoming_release = Version(major=3, minor=CURRENT_PYTHON_3_VERSION + 1)
    return (
        not v.prefix.startswith(('PyPy', 'Jython')) and v.prefix != 'nightly'
        and '-dev' not in v.suffix
        and '-alpha' not in v.suffix
        and '-beta' not in v.suffix
        and '-rc' not in v.suffix
        and v != upcoming_release
    )


def important(versions: Collection[Version]) -> VersionSet:
    """Filter out unimportant versions.

    See `is_important` for what consitutes "important".
    """
    return {
        v for v in versions
        if is_important(v)
    }


def pypy_versions(versions: Collection[Version]) -> VersionSet:
    """Filter PyPy versions."""
    return {
        v for v in versions
        if v.prefix.startswith('PyPy')
    }


def expand_pypy(versions: Collection[Version]) -> SortedVersionList:
    """Determine whether PyPy support means PyPy2 or PyPy3 or both.

    Some data sources (like setup.py classifiers) allow you to indicate PyPy
    support without specifying whether you mean PyPy2 or PyPy3.  Other data
    sources (like all CI systems) are more explicit.  To make these version
    lists directly comparable we need to look at supported CPython versions and
    translate that knowledge into PyPy versions.
    """
    supports_pypy = any(v.prefix == 'PyPy' for v in versions)
    if not supports_pypy:
        return sorted(versions)
    supports_py2 = any(v.major == 2 for v in versions)
    supports_py3 = any(v.major == 3 for v in versions)
    return sorted(
        [v for v in versions if v.prefix != 'PyPy'] +
        ([Version.from_string('PyPy')] if supports_py2 else []) +
        ([Version.from_string('PyPy3')] if supports_py3 else [])
    )


def update_version_list(
    versions: VersionList,
    add: Optional[VersionList] = None,
    drop: Optional[VersionList] = None,
    update: Optional[VersionList] = None,
) -> SortedVersionList:
    """Compute a new list of supported versions.

    ``add`` will add to supported versions.
    ``drop`` will remove from supported versions.
    ``update`` will specify supported versions.

    You may combine ``add`` and ``drop``.  It doesn't make sense to combine
    ``update`` with either ``add`` or ``drop``.
    """
    if update:
        return sorted(update)
    else:
        return sorted(set(versions).union(add or ()).difference(drop or ()))



from .appveyor import Appveyor
from .github import GitHubActions
from .manylinux import Manylinux
from .setup_py import SetupClassifiers, SetupPythonRequires
from .tox import Tox
from .travis import Travis


# The order here is only mildly important: it's used for presentation.
# Note that SetupPythonRequires.title assumes it's included right after
# SetupClassifiers!
ALL_SOURCES = [
    SetupClassifiers,
    SetupPythonRequires,
    Tox,
    Travis,
    GitHubActions,
    Appveyor,
    Manylinux,
]



import textwrap
from io import StringIO
from typing import List

import pytest

from check_python_versions.sources.tox import (
    brace_expand,
    get_tox_ini_python_versions,
    parse_envlist,
    tox_env_to_py_version,
    update_tox_envlist,
    update_tox_ini_python_versions,
)
from check_python_versions.versions import Version


def v(versions: List[str]) -> List[Version]:
    return [Version.from_string(v) for v in versions]


def test_get_tox_ini_python_versions(tmp_path):
    tox_ini = tmp_path / "tox.ini"
    tox_ini.write_text(textwrap.dedent("""\
        [tox]
        envlist = py27,py36,py27-docs,pylint,py310
    """))
    assert get_tox_ini_python_versions(tox_ini) == v(['2.7', '3.6', '3.10'])


def test_get_tox_ini_python_versions_syntax_error(tmp_path):
    tox_ini = tmp_path / "tox.ini"
    tox_ini.write_text(textwrap.dedent("""\
        ...
    """))
    assert get_tox_ini_python_versions(tox_ini) == []


def test_get_tox_ini_python_versions_no_tox_section(tmp_path):
    tox_ini = tmp_path / "tox.ini"
    tox_ini.write_text(textwrap.dedent("""\
        [flake8]
        source = foo
    """))
    assert get_tox_ini_python_versions(tox_ini) == []


def test_get_tox_ini_python_versions_no_tox_envlist(tmp_path):
    tox_ini = tmp_path / "tox.ini"
    tox_ini.write_text(textwrap.dedent("""\
        [tox]
        minversion = 3.4.0
    """))
    assert get_tox_ini_python_versions(tox_ini) == []


@pytest.mark.parametrize('s, expected', [
    ('', []),
    ('py36,py37', ['py36', 'py37']),
    ('py36, py37', ['py36', 'py37']),
    ('\n  py36,\n  py37', ['py36', 'py37']),
    ('py3{6,7},pypy', ['py36', 'py37', 'pypy']),
    ('py{35, 36, 37, 38, py3}', ['py35', 'py36', 'py37', 'py38', 'pypy3']),
])
def test_parse_envlist(s, expected):
    assert parse_envlist(s) == expected


@pytest.mark.parametrize('s, expected', [
    ('', ['']),
    ('py36', ['py36']),
    ('py3{6,7}', ['py36', 'py37']),
    ('py3{6,7}-lint', ['py36-lint', 'py37-lint']),
    ('py3{6,7}{,-lint}', ['py36', 'py36-lint', 'py37', 'py37-lint']),
    ('py3{6, 7}', ['py36', 'py37']),
    ('py3{6 ,7}', ['py36', 'py37']),
    ('py3{6 , 7}', ['py36', 'py37']),
])
def test_brace_expand(s, expected):
    assert brace_expand(s) == expected


@pytest.mark.parametrize('s, expected', [
    ('py36', '3.6'),
    ('py37-lint', '3.7'),
    ('py310', '3.10'),  # future-proofness!
    ('pypy', 'PyPy'),
    ('pypy3', 'PyPy3'),
    ('pylint', None),
    ('flake8', None),
])
def test_tox_env_to_py_version(s, expected):
    expected = expected and Version.from_string(expected)
    assert tox_env_to_py_version(s) == expected


def test_update_tox_ini_python_versions():
    fp = StringIO(textwrap.dedent("""\
        [tox]
        envlist = py26, py27
    """))
    fp.name = 'tox.ini'
    result = update_tox_ini_python_versions(fp, v(['3.6', '3.7', '3.10']))
    assert "".join(result) == textwrap.dedent("""\
        [tox]
        envlist = py36, py37, py310
    """)


def test_update_tox_ini_python_syntax_error(capsys):
    fp = StringIO(textwrap.dedent("""\
        [tox
        envlist = py26, py27
    """))
    fp.name = 'tox.ini'
    result = update_tox_ini_python_versions(fp, v(['3.6', '3.7']))
    assert "".join(result) == textwrap.dedent("""\
        [tox
        envlist = py26, py27
    """)
    assert (
        "Could not parse tox.ini:"
        in capsys.readouterr().err
    )


def test_update_tox_envlist():
    result = update_tox_envlist('docs,py26,py27,pypy3,flake8',
                                v(['3.6', '3.7']))
    assert result == 'docs,py36,py37,pypy3,flake8'


@pytest.mark.parametrize('versions, expected', [
    (['2.7', '3.7'], 'py27,py37,pypy,pypy3,flake8'),
    (['3.7'], 'py37,pypy3,flake8'),
    (['2.7'], 'py27,pypy,flake8'),
])
def test_update_tox_envlist_keeps_the_right_pypy(versions, expected):
    result = update_tox_envlist('py27,py37,pypy,pypy3,flake8', v(versions))
    assert result == expected


def test_update_tox_envlist_with_suffixes():
    result = update_tox_envlist(
        'py27,py34,py35,py36,py37,py27-numpy,py37-numpy,pypy,pypy3',
        v(['3.6', '3.7']))
    assert result == 'py36,py37,py37-numpy,pypy3'


def test_update_tox_envlist_with_spaces():
    result = update_tox_envlist(
        'py27, py34, py35, pypy3',
        v(['3.6', '3.7']))
    assert result == 'py36, py37, pypy3'


@pytest.mark.parametrize('s, expected', [
    # note that configparser trims leading whitespace, so \n is never
    # followed by a space
    ('py27,\npy34,\npy35,\npypy3', 'py36,\npy37,\npypy3'),
    ('py27\npy34\npy35\npypy3', 'py36\npy37\npypy3'),
    # uhh regression I guess
    ('\npy27,py34,py35,pypy3', 'py36,py37,pypy3'),
    # interesting corner case
    ('py27,\npy34,\npy35,\npypy3,', 'py36,\npy37,\npypy3,'),
])
def test_update_tox_envlist_with_newlines(s, expected):
    result = update_tox_envlist(s, v(['3.6', '3.7']))
    assert result == expected


@pytest.mark.parametrize('s, expected', [
    # these are contrived
    ('py{27,34,35,36,37}{,-foo,-bar},docs', 'py{36,37}{,-foo,-bar},docs'),
    ('py{27,34,py,py3}', 'py{36,37,py3}'),
    ('py27,py30,py{ramid,gmalion}', 'py36,py37,py{ramid,gmalion}'),
    ('py{27,34,py,py3},py{27,34}-docs', 'py{36,37,py3},py{36,37}-docs'),
    ('py{27,36,27-extra,36-docs}', 'py36,py37,py36-docs'),
    # these were taken from tox's documentation at
    # https://tox.readthedocs.io/en/latest/config.html#generative-envlist
    ('{py36,py27}-django{15,16}', '{py36,py37}-django{15,16}'),
    ('{py27,py36}-django{ 15, 16 }, docs, flake',
     '{py36,py37}-django{ 15, 16 }, docs, flake'),
    # these are examples from real projects
    ('py{27,py,34,35,36}-{test,stylecheck}', 'py{36,37}-{test,stylecheck}'),
    # pyjwt
    ('lint\n'
     'typing\n'
     'py{35,36,37,38}-crypto\n'
     'py{35,36,37,38}-contrib_crypto\n'
     'py{35,36,37,38}-nocrypto',
     'lint\n'
     'typing\n'
     'py{36,37}-crypto\n'
     'py{36,37}-contrib_crypto\n'
     'py{36,37}-nocrypto'),
])
def test_update_tox_envlist_with_braces(s, expected):
    result = update_tox_envlist(s, v(['3.6', '3.7']))
    assert result == expected



from typing import Callable, Optional

from ..utils import FileLines, FileOrFilename
from ..versions import SortedVersionList


ExtractorFn = Callable[[FileOrFilename], Optional[SortedVersionList]]
UpdaterFn = Callable[[FileOrFilename, SortedVersionList], Optional[FileLines]]


class Source:
    """Source for information about supported Pythons.

    Examples of sources are: setup.py, CI configuration files.

    A source has a ``title``, a typical ``filename`` (which can be a
    glob pattern).

    A source knows how to extract the list of supported Python versions
    from a file, and it can know how to update the list of versions.

    A source knows whether it's possible/typical to include PyPy support
    for packages that want to support PyPy.
    """

    def __init__(
        self,
        *,
        title: Optional[str] = None,
        filename: str,
        extract: ExtractorFn,
        update: Optional[UpdaterFn] = None,
        check_pypy_consistency: bool,
        has_upper_bound: bool,
    ) -> None:
        self.title = title or filename
        self.filename = filename
        self.extract = extract
        self.update = update
        self.check_pypy_consistency = check_pypy_consistency
        self.has_upper_bound = has_upper_bound

    def for_file(
        self,
        pathname: str,
        versions: SortedVersionList,
        relpath: str,
    ) -> 'SourceFile':
        title = relpath if self.title == self.filename else self.title
        source = SourceFile(
            title=title,
            filename=self.filename,
            extract=self.extract,
            update=self.update,
            check_pypy_consistency=self.check_pypy_consistency,
            has_upper_bound=self.has_upper_bound,
            pathname=pathname,
            versions=versions,
        )
        return source


class SourceFile(Source):
    """A concrete source file with  information about supported Pythons.

    Some sources (GitHub Actions) can have multiple files matching a
    glob pattern.  Each of those gets its own ``SourceFile`` instance.
    """

    def __init__(
        self,
        *,
        title: Optional[str] = None,
        filename: str,
        extract: ExtractorFn,
        update: Optional[UpdaterFn] = None,
        check_pypy_consistency: bool,
        has_upper_bound: bool,
        pathname: str,
        versions: SortedVersionList,
    ) -> None:
        super().__init__(
            title=title, filename=filename, extract=extract, update=update,
            check_pypy_consistency=check_pypy_consistency,
            has_upper_bound=has_upper_bound)
        self.pathname = pathname
        self.versions = versions



"""
Support for Travis CI.

Travis CI is a hosted Continuous Integration solution that can be configured
by dropping a file named ``.travis.yml`` into your source repository.

There are multiple ways of selecting Python versions, some more canonical
than others:

- via the top-level ``python`` list
- via ``python`` attributes in the jobs defined by ``jobs.include`` or its
  deprecated alias ``matrix.include``
- via ``TOXENV`` environment variables in the top-level ``env`` list
  (this is discouraged and check-python-versions might drop support for this in
  the future)
"""

from typing import Dict, List, Union

import yaml

from .base import Source
from .tox import parse_envlist, tox_env_to_py_version
from ..parsers.yaml import drop_yaml_node, quote_string, update_yaml_list
from ..utils import FileLines, FileOrFilename, open_file
from ..versions import SortedVersionList, Version, is_important


TRAVIS_YML = '.travis.yml'


# Back in the day you could do
#
#   dist: trusty
#   python:
#     - pypy
#     - pypy3
#
# but then xenial came out and it did not recognize 'pypy' or 'pypy3', instead
# requiring you to explicitly spell out full version numbers like
#
#   dist: trusty
#   python:
#     - pypy2.7-6.0.0
#     - pypy3.5-6.0.0
#
# and check-python-versions could upgrade your .travis.yml from the old version
# to the new.  Happily, this is no longer necessary, because Travis supports
# 'pypy' and 'pypy3' once again.
XENIAL_SUPPORTED_PYPY_VERSIONS: Dict[str, str] = {
    # e.g. 'pypy': 'pypy2.7-7.1.1',
}


def get_travis_yml_python_versions(
    filename: FileOrFilename = TRAVIS_YML,
) -> SortedVersionList:
    """Extract supported Python versions from .travis.yml."""
    with open_file(filename) as fp:
        conf = yaml.safe_load(fp)
    versions: List[Version] = []
    if conf.get('python'):
        if isinstance(conf['python'], list):
            versions += map(travis_normalize_py_version, conf['python'])
        else:
            versions.append(travis_normalize_py_version(conf['python']))
    if 'matrix' in conf and 'include' in conf['matrix']:
        for job in conf['matrix']['include']:
            if 'python' in job:
                versions.append(travis_normalize_py_version(job['python']))
    if 'jobs' in conf and 'include' in conf['jobs']:
        for job in conf['jobs']['include']:
            if 'python' in job:
                versions.append(travis_normalize_py_version(job['python']))
    if 'env' in conf:
        toxenvs = []
        for env in conf['env']:
            if env.startswith('TOXENV='):
                toxenvs.extend(parse_envlist(env.partition('=')[-1]))
        versions.extend(e for e in map(tox_env_to_py_version, toxenvs) if e)
    return sorted(set(versions))


def travis_normalize_py_version(v: Union[str, float]) -> Version:
    """Determine Python version from Travis ``python`` value."""
    v = str(v)
    if v.startswith('pypy3'):
        # could be pypy3, pypy3.5, pypy3.5-5.10.0
        return Version.from_string('PyPy3')
    elif v.startswith('pypy'):
        # could be pypy, pypy2, pypy2.7, pypy2.7-5.10.0
        return Version.from_string('PyPy')
    else:
        return Version.from_string(v)


def needs_xenial(v: Version) -> bool:
    """Check if a Python version needs dist: xenial.

    This is obsolete now that dist: xenial is the default, but it may
    be helpful to determine when we need to drop old dist: trusty.
    """
    return v >= Version(major=3, minor=7)


def update_travis_yml_python_versions(
    filename: FileOrFilename,
    new_versions: SortedVersionList,
) -> FileLines:
    """Update supported Python versions in .travis.yml.

    Does not touch the file but returns a list of lines with new file contents.
    """
    with open_file(filename) as fp:
        orig_lines = fp.readlines()
        fp.seek(0)
        conf = yaml.safe_load(fp)
    new_lines = orig_lines

    # Make sure we're using dist: xenial if we want to use Python 3.7 or newer.
    replacements = {}
    if any(map(needs_xenial, new_versions)):
        replacements.update(XENIAL_SUPPORTED_PYPY_VERSIONS)
        if conf.get('dist') == 'trusty':
            new_lines = drop_yaml_node(new_lines, 'dist', filename=fp.name)
        if conf.get('sudo') is False:
            # sudo is ignored nowadays, but in earlier times
            # you needed both dist: xenial and sudo: required
            # to get Python 3.7
            new_lines = drop_yaml_node(new_lines, "sudo", filename=fp.name)

    def keep_old(value: str) -> bool:
        """Determine if a Python version line should be preserved."""
        ver = travis_normalize_py_version(value)
        if ver == Version.from_string('PyPy'):
            return any(v.major == 2 for v in new_versions)
        if ver == Version.from_string('PyPy3'):
            return any(v.major == 3 for v in new_versions)
        return not is_important(ver)

    def keep_old_job(job: str) -> bool:
        """Determine if a job line should be preserved."""
        if job.startswith('python:'):
            ver = job[len('python:'):].strip()
            return not is_important(travis_normalize_py_version(ver))
        else:
            return True

    quote_style = ''
    old_versions = conf.get('python', [])
    if isinstance(old_versions, (str, int, float)):
        old_versions = [old_versions]
    for toplevel in 'matrix', 'jobs':
        for job in conf.get(toplevel, {}).get('include', []):
            if 'python' in job:
                old_versions.append(job['python'])
    if old_versions and all(isinstance(v, str) for v in old_versions):
        quote_style = '"'

    yaml_new_versions = [
        quote_string(str(v), quote_style)
        for v in new_versions
    ]

    if conf.get('python'):
        new_lines = update_yaml_list(
            new_lines, "python", yaml_new_versions, filename=fp.name,
            keep=keep_old,
            replacements=replacements,
        )
    else:
        replacements = {
            f'python: {k}': f'python: {v}'
            for k, v in replacements.items()
        }
        for toplevel in 'matrix', 'jobs':
            if 'include' not in conf.get(toplevel, {}):
                continue
            new_jobs = [
                f'python: {ver}'
                for ver in yaml_new_versions
            ]
            new_lines = update_yaml_list(
                new_lines, (toplevel, "include"), new_jobs, filename=fp.name,
                replacements=replacements, keep=keep_old_job,
            )

    # If python 3.7 was enabled via matrix.include, we've just added a
    # second 3.7 entry directly to top-level python by the above code.
    # So let's drop the matrix.

    if (
        conf.get('python')
            and 'include' in conf.get('matrix', {})
            and all(
                job.get('dist') == 'xenial'
                and set(job) <= {'python', 'dist', 'sudo'}
                for job in conf['matrix']['include']
            )
    ):
        # XXX: this may drop too much or too little!
        new_lines = drop_yaml_node(new_lines, "matrix", filename=fp.name)

    return new_lines


Travis = Source(
    filename=TRAVIS_YML,
    extract=get_travis_yml_python_versions,
    update=update_travis_yml_python_versions,
    check_pypy_consistency=True,
    has_upper_bound=True,
)



"""
Tools for manipulating Python files.
"""

import ast
import re
import string
from typing import List, Optional, Tuple, Union

from ..utils import FileLines, OneOrMore, get_indent, warn


AstValue = Union[str, List[str], Tuple[str, ...]]


def to_literal(value: str, quote_style: str = '"') -> str:
    """Convert a string value to a Python string literal."""
    # Because I don't want to deal with quoting, I'll require all values
    # to contain only safe characters (i.e. no ' or " or \).  Except some
    # PyPI classifiers do include ' so I need to handle that at least.
    # And python_requires uses all sorts of comparisons like ~= 3.7.*
    safe_chars = string.ascii_letters + string.digits + " .:,-=><!~*()/+'#"
    assert all(
        c in safe_chars for c in value
    ), f'{value!r} has unexpected characters'
    if quote_style == "'" and quote_style in value:
        quote_style = '"'
    assert quote_style not in value
    return f'{quote_style}{value}{quote_style}'


def update_call_arg_in_source(
    source_lines: FileLines,
    function: OneOrMore[str],
    keyword: str,
    new_value: Union[str, List[str]],
) -> FileLines:
    """Update a function call statement in Python source file.

    Finds the first function all, removes the old value passed to a named
    keyword argument and replaces it with a new value.

    Tries to preserve existing formatting.

    Returns the updated source.
    """
    if isinstance(function, str):
        function = (function, )
    lines = iter(enumerate(source_lines))
    rx = re.compile(fr'^({"|".join(map(re.escape, function))})\s*\(')
    for n, line in lines:
        m = rx.match(line)
        if m:
            fname = m.group(1)
            break
    else:
        warn(f'Did not find {function[0]}() call')
        return source_lines
    eq = '='
    rx = re.compile(
        f'^(?P<indent>\\s*){re.escape(keyword)}(?P<eq>\\s*=\\s*)(?P<rest>.*)'
    )
    for n, line in lines:
        m = rx.match(line)
        if m:
            first_match = m
            eq = m.group('eq')
            first_indent = m.group('indent')
            break
    else:
        warn(f'Did not find {keyword}= argument in {fname}() call')
        return source_lines

    quote_style = '"'
    rest = first_match.group('rest')
    joined = False

    if isinstance(new_value, list):
        start = n
        indent = first_indent + ' ' * 4
        if rest.startswith('[]'):
            fix_closing_bracket = True
            end = n + 1
        else:
            must_fix_indents = rest.rstrip() != '['
            fix_closing_bracket = False
            for n, line in lines:
                stripped = line.lstrip()
                if stripped.startswith(']'):
                    end = n
                    break
                elif stripped:
                    if not must_fix_indents:
                        indent = get_indent(line)
                    if stripped[0] in ('"', "'"):
                        quote_style = stripped[0]
                    if line.rstrip().endswith('],'):
                        end = n + 1
                        fix_closing_bracket = True
                        break
            else:
                warn(
                    f'Did not understand {keyword}= formatting'
                    f' in {fname}() call'
                )
                return source_lines
    elif rest.endswith('.join(['):
        joined = True
        start = n
        indent = first_indent + ' ' * 4
        for n, line in lines:
            stripped = line.lstrip()
            if stripped.startswith(']'):
                end = n + 1
                fix_closing_bracket = True
                break
        else:
            warn(
                f'Did not understand {keyword}= formatting'
                f' in {fname}() call'
            )
            return source_lines
    else:
        start = n
        end = n + 1

    if isinstance(new_value, list):
        return source_lines[:start] + [
            f"{first_indent}{keyword}{eq}[\n"
        ] + [
            f"{indent}{to_literal(value, quote_style)},\n"
            for value in new_value
        ] + ([
            f"{first_indent}],\n"
        ] if fix_closing_bracket else [
        ]) + source_lines[end:]
    elif joined:
        if rest.startswith("'"):
            quote_style = "'"
        comma = ', '
        if comma not in new_value:
            comma = ','
        new_value = new_value.split(comma)
        comma = to_literal(comma, quote_style)
        return source_lines[:start] + [
            f"{first_indent}{keyword}{eq}{comma}.join([\n"
        ] + [
            f"{indent}{to_literal(value, quote_style)},\n"
            for value in new_value
        ] + ([
            f"{first_indent}]),\n"
        ] if fix_closing_bracket else [
        ]) + source_lines[end:]
    else:
        if rest.startswith("'"):
            quote_style = "'"
        new_value_quoted = to_literal(new_value, quote_style)
        return source_lines[:start] + [
            f"{first_indent}{keyword}{eq}{new_value_quoted},\n"
        ] + source_lines[end:]


def find_call_kwarg_in_ast(
    tree: ast.AST,
    funcname: OneOrMore[str],
    keyword: str,
    *,
    filename: str,
) -> Optional[ast.AST]:
    """Find the value passed to a function call.

    ``filename`` is used for error reporting.
    """
    if isinstance(funcname, str):
        funcname = (funcname, )
    for node in ast.walk(tree):
        if (isinstance(node, ast.Call)
                and any(name_matches(n, node.func) for n in funcname)):
            for kwarg in node.keywords:
                if kwarg.arg == keyword:
                    return kwarg.value
            else:
                return None
    else:
        warn(f'Could not find {funcname[0]}() call in {filename}')
        return None


def name_matches(funcname: str, node: ast.AST) -> bool:
    """Check if the AST node refers to a funcion named `funcname`."""
    while '.' in funcname:
        funcname, dot, attr = funcname.rpartition('.')
        if not isinstance(node, ast.Attribute) or node.attr != attr:
            return False
        node = node.value
    return isinstance(node, ast.Name) and node.id == funcname


def eval_ast_node(node: ast.AST, keyword: str) -> Optional[AstValue]:
    """Partially evaluate an AST node.

    ``keyword`` is used for error reporting.
    """
    if isinstance(node, ast.Str):
        # The assert is needed to placate mypy on Python 3.8
        # https://github.com/python/mypy/issues/8837
        assert isinstance(node.s, str)
        return node.s
    if isinstance(node, (ast.List, ast.Tuple)):
        values: List[str] = []
        warned = False
        for element in node.elts:
            try:
                value = ast.literal_eval(element)
                if not isinstance(value, str):
                    raise ValueError
            except ValueError:
                pass
            else:
                values.append(value)
                continue
            if not warned:
                warn(f'Non-literal {keyword}= passed to setup(),'
                     ' skipping some values')
                warned = True
        if warned and not values:
            # no strings inside!
            return None
        if isinstance(node, ast.Tuple):
            return tuple(values)
        return values
    if (isinstance(node, ast.Call) and isinstance(node.func, ast.Attribute)
            and isinstance(node.func.value, ast.Str)
            and node.func.attr == 'join'):
        try:
            # The assert is needed to placate mypy on Python 3.8
            # https://github.com/python/mypy/issues/8837
            assert isinstance(node.func.value.s, str)
            return node.func.value.s.join(ast.literal_eval(node.args[0]))
        except ValueError:
            pass
    if isinstance(node, ast.BinOp) and isinstance(node.op, ast.Add):
        left = eval_ast_node(node.left, keyword)
        right = eval_ast_node(node.right, keyword)
        if left is not None and right is not None:
            if type(left) != type(right):
                warn(f'{keyword}= is computed by adding incompatible types:'
                     f' {type(left).__name__} and {type(right).__name__}')
                return None
            # Not sure how to make mypy accept this:
            # https://github.com/python/mypy/issues/8831
            return left + right  # type: ignore
        if left is None and right is not None:
            return right
        if left is not None and right is None:
            return left
    warn(f'Non-literal {keyword}= passed to setup()')
    return None



"""
Tools for manipulating INI files.

I want to preserve formatting and comments, therefore I cannot use a standard
INI parser and serializer.
"""

import re
from typing import List

from ..utils import FileLines, get_indent, warn


def update_ini_setting(
    orig_lines: FileLines,
    section: str,
    key: str,
    new_value: str,
    *,
    filename: str,
) -> FileLines:
    """Update a setting in an .ini file.

    Preserves formatting and comments.

    ``orig_lines`` contains the old contents of the INI file.

    ``section`` and ``key`` specify which value in which section need to be
    updated.  It is an error if the section or the key do not exist.

    ``filename`` is used for error reporting.

    Returns the updated contents.
    """
    lines = iter(enumerate(orig_lines))
    for n, line in lines:
        if line.startswith(f'[{section}]'):
            break
    else:
        warn(f'Did not find [{section}] section in {filename}')
        return orig_lines

    space = prefix = ' '
    for n, line in lines:
        m = re.match(fr'{re.escape(key)}(\s*)=(\s*)', line.rstrip())
        if m:
            start = n
            space = m.group(1)
            if not line.rstrip().endswith('='):
                prefix = m.group(2)
            break
    else:
        warn(f'Did not find {key}= in [{section}] in {filename}')
        return orig_lines

    end = start + 1
    comments = []
    pending_comments = []  # type: List[str]
    indent = '  '
    for n, line in lines:
        if line.startswith(' '):
            indent = get_indent(line)
            comments += pending_comments
            pending_comments = []
            end = n + 1
        elif line.lstrip().startswith('#'):
            pending_comments.append(line)
        else:
            break

    firstline = orig_lines[start].strip().expandtabs().replace(' ', '')
    if firstline == f'{key}=':
        if end > start + 1:
            prefix = f'\n{"".join(comments)}{indent}'

    new_value = new_value.replace('\n', '\n' + indent)
    new_lines = orig_lines[:start] + (
        f"{key}{space}={prefix}{new_value}\n"
    ).splitlines(True) + orig_lines[end:]

    return new_lines



import argparse
import os
import re
import sys
import textwrap
from io import StringIO
from typing import List

import pytest

import check_python_versions.cli as cpv
from check_python_versions.sources.base import SourceFile
from check_python_versions.versions import Version


def v(versions: List[str]) -> List[Version]:
    return [Version.from_string(v) for v in versions]


def test_parse_version_list():
    assert cpv.parse_version_list(
        '2.7,3.4-3.6'
    ) == v(['2.7', '3.4', '3.5', '3.6'])


def test_parse_version_list_3_10():
    assert cpv.parse_version_list(
        '3.7-3.10'
    ) == v(['3.7', '3.8', '3.9', '3.10'])


def test_parse_version_list_magic_range(fix_max_python_3_version):
    fix_max_python_3_version(7)
    assert cpv.parse_version_list(
        '2.7,3.4-'
    ) == v(['2.7', '3.4', '3.5', '3.6', '3.7'])
    assert cpv.parse_version_list(
        '2.6,-3.4'
    ) == v(['2.6', '3.0', '3.1', '3.2', '3.3', '3.4'])


@pytest.mark.parametrize('v', [
    '4.1-',     # unknown major version
    '-',        # both endpoints missing
    '2.7-3.4',  # major versions differ
])
def test_parse_version_list_bad_range(v):
    with pytest.raises(argparse.ArgumentTypeError,
                       match=re.escape(f'bad range: {v}')):
        cpv.parse_version_list(v)


def test_parse_version_list_bad_number():
    with pytest.raises(argparse.ArgumentTypeError):
        cpv.parse_version_list('2.x')


def test_parse_version_list_too_few():
    with pytest.raises(argparse.ArgumentTypeError):
        cpv.parse_version_list('2')


def test_parse_version_list_too_many_dots():
    with pytest.raises(argparse.ArgumentTypeError):
        cpv.parse_version_list('2.7.1')


def test_is_package(tmp_path):
    (tmp_path / "setup.py").write_text("")
    assert cpv.is_package(tmp_path)


def test_is_package_no_setup_py(tmp_path):
    assert not cpv.is_package(tmp_path)


def test_check_not_a_directory(tmp_path, capsys):
    assert not cpv.check_package(tmp_path / "xyzzy")
    assert capsys.readouterr().out == 'not a directory\n'


def test_check_not_a_package(tmp_path, capsys):
    assert not cpv.check_package(tmp_path)
    assert capsys.readouterr().out == 'no setup.py -- not a Python package?\n'


def test_check_package(tmp_path):
    setup_py = tmp_path / "setup.py"
    setup_py.write_text(textwrap.dedent("""\
        from setuptools import setup
        setup(
            name='foo',
        )
    """))
    assert cpv.check_package(tmp_path) is True


def test_check_unknown(tmp_path, capsys):
    setup_py = tmp_path / "setup.py"
    setup_py.write_text(textwrap.dedent("""\
        from setuptools import setup
        setup(
            name='foo',
        )
    """))
    assert cpv.check_versions(tmp_path) is True
    assert capsys.readouterr().out == textwrap.dedent("""\
        setup.py says: (empty)
    """)


def test_check_minimal(tmp_path, capsys):
    setup_py = tmp_path / "setup.py"
    setup_py.write_text(textwrap.dedent("""\
        from setuptools import setup
        setup(
            name='foo',
            classifiers=[
                'Programming Language :: Python :: 2.7',
                'Programming Language :: Python :: 3.6',
            ],
        )
    """))
    assert cpv.check_versions(tmp_path) is True
    assert capsys.readouterr().out == textwrap.dedent("""\
        setup.py says: 2.7, 3.6
    """)


def test_check_mismatch(tmp_path, capsys):
    setup_py = tmp_path / "setup.py"
    setup_py.write_text(textwrap.dedent("""\
        from setuptools import setup
        setup(
            name='foo',
            classifiers=[
                'Programming Language :: Python :: 2.7',
                'Programming Language :: Python :: 3.6',
            ],
        )
    """))
    tox_ini = tmp_path / "tox.ini"
    tox_ini.write_text(textwrap.dedent("""\
        [tox]
        envlist = py27
    """))
    assert cpv.check_versions(tmp_path) is False
    assert capsys.readouterr().out == textwrap.dedent("""\
        setup.py says: 2.7, 3.6
        tox.ini says:  2.7
    """)


def test_check_mismatch_pypy(tmp_path, capsys):
    setup_py = tmp_path / "setup.py"
    setup_py.write_text(textwrap.dedent("""\
        from setuptools import setup
        setup(
            name='foo',
            classifiers=[
                'Programming Language :: Python :: 2.7',
                'Programming Language :: Python :: 3.6',
                'Programming Language :: Python :: Implementation :: PyPy',
            ],
        )
    """))
    tox_ini = tmp_path / "tox.ini"
    tox_ini.write_text(textwrap.dedent("""\
        [tox]
        envlist = py27, py36, pypy
    """))
    assert cpv.check_versions(tmp_path) is False
    assert capsys.readouterr().out == textwrap.dedent("""\
        setup.py says: 2.7, 3.6, PyPy, PyPy3
        tox.ini says:  2.7, 3.6, PyPy
    """)


def test_check_expectation(tmp_path, capsys):
    setup_py = tmp_path / "setup.py"
    setup_py.write_text(textwrap.dedent("""\
        from setuptools import setup
        setup(
            name='foo',
            classifiers=[
                'Programming Language :: Python :: 2.7',
                'Programming Language :: Python :: 3.6',
                'Programming Language :: Python :: Implementation :: PyPy',
            ],
        )
    """))
    assert not cpv.check_versions(tmp_path, expect=v(['2.7', '3.6', '3.7']))
    assert capsys.readouterr().out == textwrap.dedent("""\
        setup.py says: 2.7, 3.6, PyPy, PyPy3
        expected:      2.7, 3.6, 3.7
    """)


def test_check_only(tmp_path, capsys):
    setup_py = tmp_path / "setup.py"
    setup_py.write_text(textwrap.dedent("""\
        from setuptools import setup
        setup(
            name='foo',
            classifiers=[
                'Programming Language :: Python :: 2.7',
                'Programming Language :: Python :: 3.6',
            ],
        )
    """))
    tox_ini = tmp_path / "tox.ini"
    tox_ini.write_text(textwrap.dedent("""\
        [tox]
        envlist = py27
    """))
    assert cpv.check_versions(tmp_path, only={'tox.ini'})
    assert capsys.readouterr().out == textwrap.dedent("""\
        tox.ini says: 2.7
    """)


def test_check_only_glob_source(tmp_path, capsys):
    setup_py = tmp_path / "setup.py"
    setup_py.write_text(textwrap.dedent("""\
        from setuptools import setup
        setup(
            name='foo',
            classifiers=[
                'Programming Language :: Python :: 2.7',
                'Programming Language :: Python :: 3.6',
            ],
        )
    """))
    workflow1_yml = tmp_path / ".github/workflows/one.yml"
    workflow1_yml.parent.mkdir(parents=True)
    workflow1_yml.write_text(textwrap.dedent("""\
        jobs:
          test:
            strategy:
              matrix:
                python-version: [3.6]
    """))
    assert cpv.check_versions(tmp_path, only={
        os.path.join('.github', 'workflows', 'one.yml')
    })
    result = capsys.readouterr().out.replace(os.path.sep, '/')
    assert result == textwrap.dedent("""\
        .github/workflows/one.yml says: 3.6
    """)


def test_check_only_glob(tmp_path, capsys):
    setup_py = tmp_path / "setup.py"
    setup_py.write_text(textwrap.dedent("""\
        from setuptools import setup
        setup(
            name='foo',
            classifiers=[
                'Programming Language :: Python :: 2.7',
                'Programming Language :: Python :: 3.6',
            ],
        )
    """))
    workflow1_yml = tmp_path / ".github/workflows/one.yml"
    workflow1_yml.parent.mkdir(parents=True)
    workflow1_yml.write_text(textwrap.dedent("""\
        jobs:
          test:
            strategy:
              matrix:
                python-version: [3.6]
    """))
    assert cpv.check_versions(tmp_path, only={'.github/workflows/*.yml'})
    result = capsys.readouterr().out.replace(os.path.sep, '/')
    assert result == textwrap.dedent("""\
        .github/workflows/one.yml says: 3.6
    """)


def test_check_only_nothing_matches(tmp_path, capsys):
    setup_py = tmp_path / "setup.py"
    setup_py.write_text(textwrap.dedent("""\
        from setuptools import setup
        setup(
            name='foo',
            classifiers=[
                'Programming Language :: Python :: 2.7',
                'Programming Language :: Python :: 3.6',
            ],
        )
    """))
    assert not cpv.check_versions(tmp_path, only={'nosuchfile'})
    assert capsys.readouterr().out == textwrap.dedent("""\
        no file with version information found
    """)


def test_supported_versions_match_ignores_python_requires_upper_bound():
    sources = [
        SourceFile(title='setup.py', filename='setup.py', extract=None,
                   check_pypy_consistency=True, has_upper_bound=True,
                   pathname='setup.py', versions=v(['2.7', '3.6'])),
        SourceFile(title='python_requires', filename='setup.py', extract=None,
                   check_pypy_consistency=True, has_upper_bound=False,
                   pathname='setup.py', versions=v(['2.7', '3.6', '3.7'])),
    ]
    assert cpv.supported_versions_match(sources)


def test_update_versions(tmp_path, monkeypatch):
    monkeypatch.setattr(sys, 'stdin', StringIO('y\n'))
    (tmp_path / "setup.py").write_text(textwrap.dedent("""\
        from setuptools import setup
        setup(
            name='foo',
            classifiers=[
                'Programming Language :: Python :: 2.7',
                'Programming Language :: Python :: 3.6',
            ],
        )
    """))
    cpv.update_versions(tmp_path, add=v(['3.7']))
    assert (tmp_path / "setup.py").read_text() == textwrap.dedent("""\
        from setuptools import setup
        setup(
            name='foo',
            classifiers=[
                'Programming Language :: Python :: 2.7',
                'Programming Language :: Python :: 3.6',
                'Programming Language :: Python :: 3.7',
            ],
        )
    """)


def test_update_versions_dry_run(tmp_path):
    (tmp_path / "setup.py").write_text(textwrap.dedent("""\
        from setuptools import setup
        setup(
            name='foo',
            classifiers=[
                'Programming Language :: Python :: 2.7',
                'Programming Language :: Python :: 3.6',
            ],
        )
    """))
    replacements = cpv.update_versions(tmp_path, add=v(['3.7']), dry_run=True)
    assert (tmp_path / "setup.py").read_text() == textwrap.dedent("""\
        from setuptools import setup
        setup(
            name='foo',
            classifiers=[
                'Programming Language :: Python :: 2.7',
                'Programming Language :: Python :: 3.6',
            ],
        )
    """)
    filename = str(tmp_path / "setup.py")
    assert "".join(replacements[filename]) == textwrap.dedent("""\
        from setuptools import setup
        setup(
            name='foo',
            classifiers=[
                'Programming Language :: Python :: 2.7',
                'Programming Language :: Python :: 3.6',
                'Programming Language :: Python :: 3.7',
            ],
        )
    """)


def test_update_versions_dry_run_two_updaters_one_file(
    tmp_path, fix_max_python_3_version,
):
    fix_max_python_3_version(7)
    (tmp_path / "setup.py").write_text(textwrap.dedent("""\
        from setuptools import setup
        setup(
            name='foo',
            python_requires='>=2.7',
            classifiers=[
                'Programming Language :: Python :: 2.7',
                'Programming Language :: Python :: 3.6',
            ],
        )
    """))
    replacements = cpv.update_versions(
        tmp_path, update=v(['2.7', '3.4', '3.5', '3.6', '3.7']), dry_run=True,
    )
    assert (tmp_path / "setup.py").read_text() == textwrap.dedent("""\
        from setuptools import setup
        setup(
            name='foo',
            python_requires='>=2.7',
            classifiers=[
                'Programming Language :: Python :: 2.7',
                'Programming Language :: Python :: 3.6',
            ],
        )
    """)
    filename = str(tmp_path / "setup.py")
    assert "".join(replacements[filename]) == textwrap.dedent("""\
        from setuptools import setup
        setup(
            name='foo',
            python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*',
            classifiers=[
                'Programming Language :: Python :: 2.7',
                'Programming Language :: Python :: 3.4',
                'Programming Language :: Python :: 3.5',
                'Programming Language :: Python :: 3.6',
                'Programming Language :: Python :: 3.7',
            ],
        )
    """)


def test_update_versions_diff(tmp_path, capsys):
    (tmp_path / "setup.py").write_text(textwrap.dedent("""\
        from setuptools import setup
        setup(
            name='foo',
            classifiers=[
                'Programming Language :: Python :: 2.7',
                'Programming Language :: Python :: 3.6',
            ],
        )
    """))
    cpv.update_versions(tmp_path, add=v(['3.7']), diff=True)
    assert (tmp_path / "setup.py").read_text() == textwrap.dedent("""\
        from setuptools import setup
        setup(
            name='foo',
            classifiers=[
                'Programming Language :: Python :: 2.7',
                'Programming Language :: Python :: 3.6',
            ],
        )
    """)
    assert (
        capsys.readouterr().out.replace(str(tmp_path) + os.path.sep, 'tmp/')
    ).expandtabs() == textwrap.dedent("""\
        --- tmp/setup.py        (original)
        +++ tmp/setup.py        (updated)
        @@ -4,5 +4,6 @@
             classifiers=[
                 'Programming Language :: Python :: 2.7',
                 'Programming Language :: Python :: 3.6',
        +        'Programming Language :: Python :: 3.7',
             ],
         )

    """)


def test_update_versions_no_change(tmp_path):
    (tmp_path / "setup.py").write_text(textwrap.dedent("""\
        from setuptools import setup
        setup(
            name='foo',
            classifiers=[
                'Programming Language :: Python :: 2.7',
                'Programming Language :: Python :: 3.6',
            ],
        )
    """))
    assert cpv.update_versions(tmp_path, add=v(['3.6'])) == {}


def test_update_versions_only(tmp_path):
    (tmp_path / "setup.py").write_text(textwrap.dedent("""\
        from setuptools import setup
        setup(
            name='foo',
            classifiers=[
                'Programming Language :: Python :: 2.7',
            ],
        )
    """))
    tox_ini = tmp_path / "tox.ini"
    tox_ini.write_text(textwrap.dedent("""\
        [tox]
        envlist = py27
    """))
    replacements = cpv.update_versions(
        tmp_path, add=v(['3.6']), only='tox.ini', dry_run=True,
    )
    assert set(replacements) == {str(tmp_path / 'tox.ini')}


def test_update_versions_computed(tmp_path):
    (tmp_path / "setup.py").write_text(textwrap.dedent("""\
        from setuptools import setup
        setup(
            name='foo',
            classifiers=[
                'Programming Language :: Python :: %s' % v
                for v in ['2.7']
            ],
        )
    """))
    replacements = cpv.update_versions(
        tmp_path, add=v(['3.6']), dry_run=True,
    )
    assert set(replacements) == set()


def test_main_help(monkeypatch):
    monkeypatch.setattr(sys, 'argv', ['check-python-versions', '--help'])
    with pytest.raises(SystemExit):
        cpv.main()


@pytest.mark.parametrize('arg', [
    'xyzzy',
    '1,2,3',
    '2.x',
    '1.2.3',
    '2.7-3.6',
])
def test_main_expect_error_handling(monkeypatch, arg, capsys):
    monkeypatch.setattr(sys, 'argv', [
        'check-python-versions', '--expect', arg,
    ])
    with pytest.raises(SystemExit):
        cpv.main()
    # the error is either 'bad version: ...' or 'bad range: ...'
    assert '--expect: bad' in capsys.readouterr().err


@pytest.mark.parametrize('arg', ['--add', '--drop'])
def test_main_conflicting_args(monkeypatch, tmp_path, capsys, arg):
    monkeypatch.setattr(sys, 'argv', [
        'check-python-versions',
        str(tmp_path),
        arg, '3.8',
        '--update', '3.6-3.7',
    ])
    with pytest.raises(SystemExit):
        cpv.main()
    assert (
        f'argument {arg}: not allowed with argument --update'
        in capsys.readouterr().err
    )


@pytest.mark.parametrize('arg', ['--diff', '--dry-run'])
def test_main_required_args(monkeypatch, tmp_path, capsys, arg):
    monkeypatch.setattr(sys, 'argv', [
        'check-python-versions',
        str(tmp_path),
        arg,
    ])
    with pytest.raises(SystemExit):
        cpv.main()
    assert (
        f'argument {arg}: not allowed without --update/--add/--drop'
        in capsys.readouterr().err
    )


def test_main_diff_and_expect_and_dry_run_oh_my(monkeypatch, tmp_path, capsys):
    monkeypatch.setattr(sys, 'argv', [
        'check-python-versions',
        str(tmp_path),
        '--expect', '3.6-3.7',
        '--update', '3.6-3.7',
        '--diff',
    ])
    with pytest.raises(SystemExit):
        cpv.main()
    assert (
        'argument --expect: not allowed with --diff,'
        ' unless you also add --dry-run'
        in capsys.readouterr().err
    )


def test_main_here(monkeypatch, capsys):
    monkeypatch.setattr(sys, 'argv', [
        'check-python-versions',
    ])
    cpv.main()
    assert 'mismatch' not in capsys.readouterr().out


def test_main_skip_non_packages(monkeypatch, capsys, tmp_path):
    monkeypatch.setattr(sys, 'argv', [
        'check-python-versions', '--skip-non-packages', str(tmp_path),
    ])
    cpv.main()
    assert capsys.readouterr().out == ''


def test_main_single(monkeypatch, capsys, tmp_path):
    monkeypatch.setattr(sys, 'argv', [
        'check-python-versions',
        str(tmp_path / "a"),
    ])
    with pytest.raises(SystemExit) as exc_info:
        cpv.main()
    assert (
        capsys.readouterr().out + str(exc_info.value) + '\n'
    ).replace(str(tmp_path), 'tmp') == textwrap.dedent("""\
        not a directory

        mismatch!
    """)


def test_main_only(monkeypatch, capsys, tmp_path):
    monkeypatch.setattr(sys, 'argv', [
        'check-python-versions',
        str(tmp_path),
        '--only', 'tox.ini,setup.py',
    ])
    setup_py = tmp_path / "setup.py"
    setup_py.write_text(textwrap.dedent("""\
        from setuptools import setup
        setup(
            name='foo',
            classifiers=[
                'Programming Language :: Python :: 2.7',
                'Programming Language :: Python :: 3.6',
            ],
        )
    """))
    tox_ini = tmp_path / "tox.ini"
    tox_ini.write_text(textwrap.dedent("""\
        [tox]
        envlist = py27,py36
    """))
    travis_yml = tmp_path / ".travis.yml"
    travis_yml.write_text(textwrap.dedent("""\
        python:
          - 2.7
          - 3.5
    """))
    cpv.main()
    assert (
        capsys.readouterr().out + '\n'
    ).replace(str(tmp_path) + os.path.sep, 'tmp/') == textwrap.dedent("""\
        setup.py says: 2.7, 3.6
        tox.ini says:  2.7, 3.6

    """)


def test_main_multiple(monkeypatch, capsys, tmp_path):
    monkeypatch.setattr(sys, 'argv', [
        'check-python-versions',
        str(tmp_path / "a"),
        str(tmp_path / "b"),
        '--expect', '3.6, 3.7'
    ])
    (tmp_path / "a").mkdir()
    (tmp_path / "a" / "setup.py").write_text(textwrap.dedent("""\
        from setuptools import setup
        setup(
            name='foo',
            classifiers=[
                'Programming Language :: Python :: 2.7',
                'Programming Language :: Python :: 3.6',
            ],
        )
    """))
    with pytest.raises(SystemExit) as exc_info:
        cpv.main()
    assert (
        capsys.readouterr().out + str(exc_info.value) + '\n'
    ).replace(str(tmp_path) + os.path.sep, 'tmp/') == textwrap.dedent("""\
        tmp/a:

        setup.py says:                2.7, 3.6
        expected:                     3.6, 3.7


        tmp/b:

        not a directory


        mismatch in tmp/a tmp/b!
    """)


def test_main_multiple_ok(monkeypatch, tmp_path, capsys):
    monkeypatch.setattr(sys, 'argv', [
        'check-python-versions',
        str(tmp_path / "a"),
        str(tmp_path / "b"),
    ])
    (tmp_path / "a").mkdir()
    (tmp_path / "a" / "setup.py").write_text(textwrap.dedent("""\
        from setuptools import setup
        setup(
            name='foo',
            classifiers=[
                'Programming Language :: Python :: 2.7',
                'Programming Language :: Python :: 3.6',
            ],
        )
    """))
    (tmp_path / "b").mkdir()
    (tmp_path / "b" / "setup.py").write_text(textwrap.dedent("""\
        from setuptools import setup
        setup(
            name='foo',
            classifiers=[
                'Programming Language :: Python :: 3.6',
                'Programming Language :: Python :: 3.7',
            ],
        )
    """))
    cpv.main()
    assert (
        capsys.readouterr().out.endswith('\n\nall ok!\n')
    )


def test_main_update(monkeypatch, capsys, tmp_path):
    monkeypatch.setattr(sys, 'stdin', StringIO('y\n'))
    monkeypatch.setattr(sys, 'argv', [
        'check-python-versions',
        str(tmp_path),
        '--add', '3.7,3.8',
    ])
    (tmp_path / "setup.py").write_text(textwrap.dedent("""\
        from setuptools import setup
        setup(
            name='foo',
            classifiers=[
                'Programming Language :: Python :: 2.7',
                'Programming Language :: Python :: 3.6',
            ],
        )
    """))
    cpv.main()
    assert (
        capsys.readouterr().out
        .replace(str(tmp_path) + os.path.sep, 'tmp/')
        .expandtabs()
        .replace(' \n', '\n\n')
    ) == textwrap.dedent("""\
        --- tmp/setup.py        (original)
        +++ tmp/setup.py        (updated)
        @@ -4,5 +4,7 @@
             classifiers=[
                 'Programming Language :: Python :: 2.7',
                 'Programming Language :: Python :: 3.6',
        +        'Programming Language :: Python :: 3.7',
        +        'Programming Language :: Python :: 3.8',
             ],
         )

        Write changes to tmp/setup.py? [y/N]

        setup.py says: 2.7, 3.6, 3.7, 3.8
    """)
    assert (tmp_path / "setup.py").read_text() == textwrap.dedent("""\
        from setuptools import setup
        setup(
            name='foo',
            classifiers=[
                'Programming Language :: Python :: 2.7',
                'Programming Language :: Python :: 3.6',
                'Programming Language :: Python :: 3.7',
                'Programming Language :: Python :: 3.8',
            ],
        )
    """)


def test_main_update_rejected(monkeypatch, capsys, tmp_path):
    monkeypatch.setattr(sys, 'stdin', StringIO('n\n'))
    monkeypatch.setattr(sys, 'argv', [
        'check-python-versions',
        str(tmp_path),
        '--add', '3.7,3.8',
    ])
    (tmp_path / "setup.py").write_text(textwrap.dedent("""\
        from setuptools import setup
        setup(
            name='foo',
            classifiers=[
                'Programming Language :: Python :: 2.7',
                'Programming Language :: Python :: 3.6',
            ],
        )
    """))
    cpv.main()
    assert (
        capsys.readouterr().out
        .replace(str(tmp_path) + os.path.sep, 'tmp/')
        .expandtabs()
        .replace(' \n', '\n\n')
    ) == textwrap.dedent("""\
        --- tmp/setup.py        (original)
        +++ tmp/setup.py        (updated)
        @@ -4,5 +4,7 @@
             classifiers=[
                 'Programming Language :: Python :: 2.7',
                 'Programming Language :: Python :: 3.6',
        +        'Programming Language :: Python :: 3.7',
        +        'Programming Language :: Python :: 3.8',
             ],
         )

        Write changes to tmp/setup.py? [y/N]

        setup.py says: 2.7, 3.6
    """)
    assert (tmp_path / "setup.py").read_text() == textwrap.dedent("""\
        from setuptools import setup
        setup(
            name='foo',
            classifiers=[
                'Programming Language :: Python :: 2.7',
                'Programming Language :: Python :: 3.6',
            ],
        )
    """)


def test_main_update_diff(monkeypatch, capsys, tmp_path):
    monkeypatch.setattr(sys, 'argv', [
        'check-python-versions',
        str(tmp_path),
        '--add', '3.7,3.8',
        '--diff',
    ])
    (tmp_path / "setup.py").write_text(textwrap.dedent("""\
        from setuptools import setup
        setup(
            name='foo',
            classifiers=[
                'Programming Language :: Python :: 2.7',
                'Programming Language :: Python :: 3.6',
            ],
        )
    """))
    cpv.main()
    assert (
        capsys.readouterr().out
        .replace(str(tmp_path) + os.path.sep, 'tmp/')
        .expandtabs()
        .replace(' \n', '\n\n')
    ) == textwrap.dedent("""\
        --- tmp/setup.py        (original)
        +++ tmp/setup.py        (updated)
        @@ -4,5 +4,7 @@
             classifiers=[
                 'Programming Language :: Python :: 2.7',
                 'Programming Language :: Python :: 3.6',
        +        'Programming Language :: Python :: 3.7',
        +        'Programming Language :: Python :: 3.8',
             ],
         )

    """)
    assert (tmp_path / "setup.py").read_text() == textwrap.dedent("""\
        from setuptools import setup
        setup(
            name='foo',
            classifiers=[
                'Programming Language :: Python :: 2.7',
                'Programming Language :: Python :: 3.6',
            ],
        )
    """)


def test_main_update_dry_run(monkeypatch, capsys, tmp_path):
    monkeypatch.setattr(sys, 'argv', [
        'check-python-versions',
        str(tmp_path),
        '--add', '3.7,3.8',
        '--dry-run',
    ])
    (tmp_path / "setup.py").write_text(textwrap.dedent("""\
        from setuptools import setup
        setup(
            name='foo',
            classifiers=[
                'Programming Language :: Python :: 2.7',
                'Programming Language :: Python :: 3.6',
            ],
        )
    """))
    cpv.main()
    assert (
        capsys.readouterr().out
        .replace(str(tmp_path) + os.path.sep, 'tmp/')
        .expandtabs()
        .replace(' \n', '\n\n')
    ) == textwrap.dedent("""\
        setup.py says: 2.7, 3.6, 3.7, 3.8
    """)
    assert (tmp_path / "setup.py").read_text() == textwrap.dedent("""\
        from setuptools import setup
        setup(
            name='foo',
            classifiers=[
                'Programming Language :: Python :: 2.7',
                'Programming Language :: Python :: 3.6',
            ],
        )
    """)


def test_main_handles_ctrl_c(monkeypatch):
    def raise_keyboard_interrupt():
        raise KeyboardInterrupt()
    monkeypatch.setattr(cpv, '_main', raise_keyboard_interrupt)
    with pytest.raises(SystemExit):
        cpv.main()
